import datetime
import errno
import json
import logging
import os
from copy import deepcopy

import click
import pytest
import yaml

from botocore.exceptions import ClientError
from click.testing import CliRunner
from deepdiff import DeepDiff
from unittest.mock import MagicMock, patch, sentinel

from sceptre.cli import cli
from sceptre.cli.helpers import CustomJsonEncoder, \
    catch_exceptions, setup_logging, write, \
    ColouredFormatter, deserialize_json_properties

from sceptre.config.reader import ConfigReader

from sceptre.diffing.stack_differ import \
    DeepDiffStackDiffer, DifflibStackDiffer, StackDiff

from sceptre.exceptions import SceptreException
from sceptre.plan.actions import StackActions
from sceptre.stack import Stack
from sceptre.stack_status import StackStatus


class TestCli:

    def setup_method(self, test_method):
        self.patcher_ConfigReader = patch("sceptre.plan.plan.ConfigReader")
        self.patcher_StackActions = patch("sceptre.plan.executor.StackActions")

        self.mock_ConfigReader = self.patcher_ConfigReader.start()
        self.mock_StackActions = self.patcher_StackActions.start()

        self.mock_config_reader = MagicMock(spec=ConfigReader)
        self.mock_stack_actions = MagicMock(spec=StackActions)

        self.mock_stack = MagicMock(
            spec=Stack,
            region=None,
            profile=None,
            external_name='mock-stack-external',
            dependencies=[],
            ignore=False,
            obsolete=False,
        )

        self.mock_stack.name = 'mock-stack'

        self.mock_config_reader.construct_stacks.return_value = \
            set([self.mock_stack]), set([self.mock_stack])

        self.mock_stack_actions.stack = self.mock_stack

        self.mock_ConfigReader.return_value = self.mock_config_reader
        self.mock_StackActions.return_value = self.mock_stack_actions

        self.runner = CliRunner()

    def teardown_method(self, test_method):
        self.patcher_ConfigReader.stop()
        self.patcher_StackActions.stop()

    @patch("sys.exit")
    def test_catch_exceptions(self, mock_exit):
        @catch_exceptions
        def raises_exception():
            raise SceptreException()

        raises_exception()
        mock_exit.assert_called_once_with(1)

    def test_catch_exceptions_debug_mode(self):
        @catch_exceptions
        def raises_exception():
            raise SceptreException()

        logger = logging.getLogger("sceptre")
        logger.setLevel(logging.DEBUG)

        with pytest.raises(SceptreException):
            raises_exception()

    @pytest.mark.parametrize("command,files,output", [
        # one --var option
        (
            ["--var", "a=1", "noop"],
            {},
            {"a": "1"}
        ),
        # multiple --var options
        (
            ["--var", "a=1", "--var", "b=2", "noop"],
            {},
            {"a": "1", "b": "2"}
        ),
        # multiple --var options same key
        (
            ["--var", "a=1", "--var", "a=2", "noop"],
            {},
            {"a": "2"}
        ),
        (
            ["--var-file", "foo.yaml", "--var", "key3.subkey1.id=id2", "noop"],
            {
                "foo.yaml": {
                    "key1": "val1",
                    "key2": "val2",
                    "key3": {
                        "subkey1": {
                            "id": "id1"
                        }
                    }
                }
            },
            {
                "key1": "val1",
                "key2": "val2",
                "key3": {"subkey1": {"id": "id2"}}
            }
        ),
        # one --var-file option
        (
            ["--var-file", "foo.yaml", "noop"],
            {
                "foo.yaml": {"key1": "val1", "key2": "val2"}
            },
            {"key1": "val1", "key2": "val2"}
        ),
        # multiple --var-file option
        (
            ["--var-file", "foo.yaml", "--var-file", "bar.yaml", "noop"],
            {
                "foo.yaml": {"key1": "parent_value1", "key2": "parent_value2"},
                "bar.yaml": {"key2": "child_value2", "key3": "child_value3"}
            },
            {
                "key1": "parent_value1",
                "key2": "child_value2",
                "key3": "child_value3"
            }
        ),
        # mix of --var and --var-file
        (
            ["--var-file", "foo.yaml", "--var", "key2=var2", "noop"],
            {
                "foo.yaml": {"key1": "file1", "key2": "file2"}
            },
            {"key1": "file1", "key2": "var2"}
        ),
        # multiple --var-file option, illustrating dictionaries not merged.
        (
            ["--var-file", "foo.yaml", "--var-file", "bar.yaml", "noop"],
            {
                "foo.yaml": {"key1": {"a": "b"}},
                "bar.yaml": {"key1": {"c": "d"}}
            },
            {
                "key1": {"c": "d"}
            }
        ),
        # multiple --var-file option, dictionaries merged.
        (
            ["--merge-vars", "--var-file", "foo.yaml", "--var-file", "bar.yaml", "noop"],
            {
                "foo.yaml": {"key1": {"a": "b"}},
                "bar.yaml": {"key1": {"c": "d"}}
            },
            {
                "key1": {"a": "b", "c": "d"}
            }
        ),
        # multiple --var-file option, dictionaries merged, complex example.
        (
            ["--merge-vars", "--var-file", "common.yaml", "--var-file", "dev.yaml", "noop"],
            {
                "common.yaml": {
                    "CommonTags": {
                        "Organization": "Parts Unlimited",
                        "Department": "IT Operations"
                    }
                },
                "dev.yaml": {"CommonTags": {"Environment": "dev"}}
            },
            {
                "CommonTags": {
                    "Organization": "Parts Unlimited",
                    "Department": "IT Operations",
                    "Environment": "dev"
                }
            }
        ),
        # multiple --var-file option, dictionaries merged, complex example, with overrides.
        (
            ["--merge-vars", "--var-file", "common.yaml", "--var-file", "dev.yaml", "noop"],
            {
                "common.yaml": {
                    "CommonTags": {
                        "Organization": "Parts Unlimited",
                        "Department": "IT Operations",
                        "Environment": "sandbox"
                    }
                },
                "dev.yaml": {"CommonTags": {"Environment": "dev"}},
            },
            {
                "CommonTags": {
                    "Organization": "Parts Unlimited",
                    "Department": "IT Operations",
                    "Environment": "dev"
                }
            }
        ),
        # multiple --var-file option, dictionaries merged, complex example, with lists.
        (
            ["--merge-vars", "--var-file", "common.yaml", "--var-file", "test.yaml", "noop"],
            {
                "common.yaml": {
                    "CommonTags": {
                        "Organization": "Parts Unlimited",
                        "Department": "IT Operations",
                        "Envlist": ["sandbox", "dev"]
                    }
                },
                "test.yaml": {"CommonTags": {"Envlist": ["test"]}},
            },
            {
                "CommonTags": {
                    "Organization": "Parts Unlimited",
                    "Department": "IT Operations",
                    "Envlist": ["test"]
                }
            }
        ),
        # multiple --var-file option, dictionaries merged, multiple levels.
        (
            ["--merge-vars", "--var-file", "common.yaml", "--var-file", "test.yaml", "noop"],
            {
                "common.yaml": {"a": {"b": {"c": "p", "d": "q"}}},
                "test.yaml": {"a": {"b": {"c": "r", "e": "s"}}}
            },
            {
                "a": {"b": {"c": "r", "d": "q", "e": "s"}}
            }
        ),
        # a --var-file and --var combined.
        (
            ["--merge-vars", "--var-file", "common.yaml", "--var", "CommonTags.Version=1.0.0", "noop"],
            {
                "common.yaml": {
                    "CommonTags": {
                        "Organization": "Parts Unlimited",
                        "Department": "IT Operations",
                        "Envlist": ["sandbox", "dev"]
                    }
                }
            },
            {
                "CommonTags": {
                    "Organization": "Parts Unlimited",
                    "Department": "IT Operations",
                    "Envlist": ["sandbox", "dev"],
                    "Version": "1.0.0"
                }
            }
        ),
        # multiple --var-file and --var combined.
        (
            [
                "--merge-vars", "--var-file", "common.yaml", "--var-file", "test.yaml",
                "--var", "CommonTags.Project=Unboxing", "noop"
            ],
            {
                "common.yaml": {
                    "CommonTags": {
                        "Organization": "Parts Unlimited",
                        "Department": "IT Operations",
                        "Envlist": ["sandbox", "dev"]
                    }
                },
                "test.yaml": {
                    "CommonTags": {
                        "Project": "Boxing"
                    }
                }
            },
            {
                "CommonTags": {
                    "Organization": "Parts Unlimited",
                    "Department": "IT Operations",
                    "Envlist": ["sandbox", "dev"],
                    "Project": "Unboxing"
                }
            }
        )
    ])
    def test_user_variables(self, command, files, output):
        @cli.command()
        @click.pass_context
        def noop(ctx):
            click.echo(yaml.safe_dump(ctx.obj.get("user_variables")))

        with self.runner.isolated_filesystem():
            for name, content in files.items():
                with open(name, "w") as fh:
                    yaml.safe_dump(content, fh)

            result = self.runner.invoke(cli, command)

        user_variables = yaml.safe_load(result.output)
        assert result.exit_code == 0
        assert user_variables == output

    def test_validate_template_with_valid_template(self):
        self.mock_stack_actions.validate.return_value = {
            "Parameters": "Example",
            "ResponseMetadata": {
                "HTTPStatusCode": 200
            }
        }

        result_json = json.dumps({'Parameters': 'Example'}, indent=4)
        result = self.runner.invoke(cli, ["--output", "json", "validate", "dev/vpc.yaml"])
        self.mock_stack_actions.validate.assert_called_with()
        assert result.output == "Template mock-stack is valid. Template details:\n\n{}\n".format(
            result_json)

    def test_validate_template_with_invalid_template(self):
        client_error = ClientError(
            {
                "Errors":
                {
                    "Message": "Unrecognized resource types",
                    "Code": "ValidationError",
                }
            },
            "ValidateTemplate"
        )
        self.mock_stack_actions.validate.side_effect = client_error

        expected_result = str(client_error) + "\n"
        result = self.runner.invoke(cli, ["--output", "json", "validate", "dev/vpc.yaml"])
        assert expected_result in result.output.replace("\"", "")

    def test_estimate_template_cost_with_browser(self):
        self.mock_stack_actions.estimate_cost.return_value = {
            "Url": "https://docs.sceptre-project.org",
            "ResponseMetadata": {
                "HTTPStatusCode": 200
            }
        }

        args = ["estimate-cost", "dev/vpc.yaml"]
        with patch('webbrowser.open', return_value=None):  # Do not open a web browser
            result = self.runner.invoke(cli, args)

        self.mock_stack_actions.estimate_cost.assert_called_with()

        assert result.output == \
            '{0}{1}'.format("View the estimated cost for mock-stack at:\n",
                            "https://docs.sceptre-project.org\n\n")

    def test_estimate_template_cost_with_no_browser(self):
        client_error = ClientError(
            {
                "Errors":
                {
                    "Message": "No Browser",
                    "Code": "Error",
                }
            },
            "Webbrowser"
        )
        self.mock_stack_actions.estimate_cost.side_effect = client_error
        expected_result = "{}\n".format(client_error)
        result = self.runner.invoke(
            cli,
            ["estimate-cost", "dev/vpc.yaml"]
        )
        assert expected_result in result.output.replace("\"", "")

    def test_lock_stack(self):
        self.runner.invoke(
            cli, ["set-policy", "dev/vpc.yaml", "-b", "deny-all"]
        )
        self.mock_config_reader.construct_stacks.assert_called_with()
        self.mock_stack_actions.lock.assert_called_with()

    def test_unlock_stack(self):
        self.runner.invoke(
            cli, ["set-policy", "dev/vpc.yaml", "-b", "allow-all"]
        )
        self.mock_config_reader.construct_stacks.assert_called_with()
        self.mock_stack_actions.unlock.assert_called_with()

    def test_set_policy_with_file_flag(self):
        policy_file = "tests/fixtures/stack_policies/lock.json"
        result = self.runner.invoke(cli, [
            "set-policy", "dev/vpc.yaml", policy_file
        ])
        assert result.exit_code == 0

    def test_describe_policy_with_existing_policy(self):
        self.mock_stack_actions.get_policy.return_value = {
            "dev/vpc": {"Statement": ["Body"]}
        }

        result = self.runner.invoke(
            cli, ["--output", "json", "describe", "policy", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert result.output == "{}\n".format(json.dumps(
            {'dev/vpc': {'Statement': ['Body']}}, indent=4))

    def test_list_group_resources(self):
        response = {
            "stack-name-1": {
                "StackResources": [
                    {
                        "LogicalResourceId": "logical-resource-id",
                        "PhysicalResourceId": "physical-resource-id"
                    }
                ]
            },
            "stack-name-2": {
                "StackResources": [
                    {
                        "LogicalResourceId": "logical-resource-id",
                        "PhysicalResourceId": "physical-resource-id"
                    }
                ]
            }
        }
        self.mock_stack_actions.describe_resources.return_value = response
        result = self.runner.invoke(cli, ["--output", "yaml", "list", "resources", "dev"])

        assert yaml.safe_load(result.output) == [response]
        assert result.exit_code == 0

    def test_list_stack_resources(self):
        response = {
            "StackResources": [
                {
                    "LogicalResourceId": "logical-resource-id",
                    "PhysicalResourceId": "physical-resource-id"
                }
            ]
        }
        self.mock_stack_actions.describe_resources.return_value = response
        result = self.runner.invoke(cli, ["--output", "yaml", "list", "resources", "dev/vpc.yaml"])
        assert yaml.safe_load(result.output) == [response]
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "command,success,yes_flag,exit_code", [
            ("create", True, True, 0),
            ("create", False, True, 1),
            ("create", True, False, 0),
            ("create", False, False, 1),
            ("delete", True, True, 0),
            ("delete", False, True, 1),
            ("delete", True, False, 0),
            ("delete", False, False, 1),
            ("update", True, True, 0),
            ("update", False, True, 1),
            ("update", True, False, 0),
            ("update", False, False, 1),
            ("launch", True, True, 0),
            ("launch", False, True, 1),
            ("launch", True, False, 0),
            ("launch", False, False, 1)
        ]
    )
    def test_stack_commands(self, command, success, yes_flag, exit_code):
        run_command = getattr(self.mock_stack_actions, command)
        run_command.return_value = \
            StackStatus.COMPLETE if success else StackStatus.FAILED

        kwargs = {"args": [command, "dev/vpc.yaml"]}
        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        run_command.assert_called_with()
        assert result.exit_code == exit_code

    @pytest.mark.parametrize(
        "command, ignore_dependencies", [
            ("create", True),
            ("create", False),
            ("delete", True),
            ("delete", False),
        ]
    )
    def test_ignore_dependencies_commands(self, command, ignore_dependencies):
        args = [command, "dev/vpc.yaml", "cs-1", "-y"]
        if ignore_dependencies:
            args.insert(0, "--ignore-dependencies")
        result = self.runner.invoke(cli, args)
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "command,yes_flag", [
            ("create", True),
            ("create", False),
            ("delete", True),
            ("delete", False),
            ("execute", True),
            ("execute", False)
        ]
    )
    def test_change_set_commands(self, command, yes_flag):
        stack_command = command + "_change_set"

        kwargs = {"args": [command, "dev/vpc.yaml", "cs1"]}
        if yes_flag:
            kwargs["args"].append("-y")
        else:
            kwargs["input"] = "y\n"

        result = self.runner.invoke(cli, **kwargs)

        getattr(self.mock_stack_actions,
                stack_command).assert_called_with("cs1")
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        "verbose_flag,", [
            (False),
            (True)
        ]
    )
    def test_describe_change_set(self, verbose_flag):
        response = {
            "VerboseProperty": "VerboseProperty",
            "ChangeSetName": "ChangeSetName",
            "CreationTime": "CreationTime",
            "ExecutionStatus": "ExecutionStatus",
            "StackName": "StackName",
            "Status": "Status",
            "StatusReason": "StatusReason",
            "Changes": [
                {
                    "ResourceChange": {
                        "Action": "Action",
                        "LogicalResourceId": "LogicalResourceId",
                        "PhysicalResourceId": "PhysicalResourceId",
                        "Replacement": "Replacement",
                        "ResourceType": "ResourceType",
                        "Scope": "Scope",
                        "VerboseProperty": "VerboseProperty"
                    }
                }
            ]
        }
        args = ["describe", "change-set", "region/vpc.yaml", "cs1"]
        if verbose_flag:
            args.append("-v")

        self.mock_stack_actions.describe_change_set.return_value = response
        result = self.runner.invoke(cli, args)
        if not verbose_flag:
            del response["VerboseProperty"]
            del response["Changes"][0]["ResourceChange"]["VerboseProperty"]
        assert yaml.safe_load(result.output) == response
        assert result.exit_code == 0

    def test_list_change_sets_with_200(self):
        self.mock_stack_actions.list_change_sets.return_value = {
            "ChangeSets": "Test"
        }
        result = self.runner.invoke(
            cli, ["list", "change-sets", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert yaml.safe_load(result.output) == {"ChangeSets": "Test"}

    def test_list_change_sets_without_200(self):
        response = {
            "ChangeSets": "Test"
        }
        self.mock_stack_actions.list_change_sets.return_value = response

        result = self.runner.invoke(
            cli, ["--output", "json", "list", "change-sets", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert yaml.safe_load(result.output) == response

    def test_list_outputs_json(self):
        outputs = {"OutputKey": "Key", "OutputValue": "Value"}
        self.mock_stack_actions.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["--output", "json", "list", "outputs", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert json.loads(result.output) == [outputs]

    def test_list_outputs_yaml(self):
        outputs = {"OutputKey": "Key", "OutputValue": "Value"}
        self.mock_stack_actions.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["--output", "yaml", "list", "outputs", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        expected_output = '---\n- OutputKey: Key\n  OutputValue: Value\n\n'
        assert result.output == expected_output

    def test_list_outputs_text(self):
        outputs = {"StackName": [{'OutputKey': "Key", "OutputValue": "Value"}]}
        self.mock_stack_actions.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["--output", "text", "list", "outputs", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        expected_output = 'StackOutputKeyOutputValue\n\nStackNameKeyValue\n'
        assert result.output.replace(' ', '') == expected_output

    def test_list_outputs_with_export(self):
        outputs = {'stack': [{'OutputKey': 'Key', 'OutputValue': 'Value'}]}
        self.mock_stack_actions.describe_outputs.return_value = outputs
        result = self.runner.invoke(
            cli, ["list", "outputs", "dev/vpc.yaml", "-e", "envvar"]
        )
        assert result.exit_code == 0
        assert result.output == "export SCEPTRE_Key='Value'\n"

    @pytest.mark.parametrize("path,output_format,expected_output", [
        ("dev/vpc.yaml", "yaml", '---\nmock-stack.yaml: mock-stack-external\n\n'),
        ("dev/vpc.yaml", "text", '---\nmock-stack.yaml: mock-stack-external\n\n'),
        ("dev/vpc.yaml", "json", '{\n    "mock-stack.yaml": "mock-stack-external"\n}\n'),
        ("dev",          "yaml", '---\nmock-stack.yaml: mock-stack-external\n\n')
    ])
    def test_list_stacks(self, path, output_format, expected_output):
        result = self.runner.invoke(
            cli, ["--output", output_format, "list", "stacks", path]
        )
        assert result.exit_code == 0
        assert result.stdout == expected_output

    def test_status_with_group(self):
        self.mock_stack_actions.get_status.return_value = {
            "stack": "status"
        }

        result = self.runner.invoke(cli, ["--output", "json", "status", "dev"])
        assert result.exit_code == 0
        assert result.output == '{\n    "mock-stack": {\n        \"stack\": \"status\"\n    }\n}\n'

    def test_status_with_stack(self):
        self.mock_stack_actions.get_status.return_value = "status"
        result = self.runner.invoke(cli, ["status", "dev/vpc.yaml"])
        assert result.exit_code == 0
        assert result.output == '{\n    "mock-stack": "status"\n}\n'

    def test_new_project_non_existant(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            template_dir = os.path.join(project_path, "templates")
            region = "test-region"
            os.environ["AWS_DEFAULT_REGION"] = region
            defaults = {
                "project_code": "example",
                "region": region
            }

            result = self.runner.invoke(cli, ["new", "project", "example"])
            assert not result.exception
            assert os.path.isdir(config_dir)
            assert os.path.isdir(template_dir)

            with open(os.path.join(config_dir, "config.yaml")) as config_file:
                config = yaml.safe_load(config_file)

            assert config == defaults

    def test_new_project_already_exist(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            template_dir = os.path.join(project_path, "templates")
            existing_config = {"Test": "Test"}

            os.mkdir(project_path)
            os.mkdir(config_dir)
            os.mkdir(template_dir)

            config_filepath = os.path.join(config_dir, "config.yaml")
            with open(config_filepath, 'w') as config_file:
                yaml.dump(existing_config, config_file)

            result = self.runner.invoke(cli, ["new", "project", "example"])
            assert result.exit_code == 1
            assert result.output == '"Folder \\"example\\" already exists."\n'
            assert os.path.isdir(config_dir)
            assert os.path.isdir(template_dir)

            with open(os.path.join(config_dir, "config.yaml")) as config_file:
                config = yaml.safe_load(config_file)
            assert existing_config == config

    def test_new_project_another_exception(self):
        with self.runner.isolated_filesystem():
            patcher_mkdir = patch("sceptre.cli.new.os.mkdir")
            mock_mkdir = patcher_mkdir.start()
            mock_mkdir.side_effect = OSError(errno.EINVAL)
            result = self.runner.invoke(cli, ["new", "project", "example"])
            mock_mkdir = patcher_mkdir.stop()
            assert str(result.exception) == str(OSError(errno.EINVAL))

    @pytest.mark.parametrize(
        "stack_group,config_structure,stdin,result", [
            (
                "A",
                {"": {}},
                'y\nA\nA\n', {"project_code": "A", "region": "A"}
            ),
            (
                "A",
                {"": {"project_code": "top", "region": "top"}},
                'y\n\n\n', {}
            ),
            (
                "A",
                {"": {"project_code": "top", "region": "top"}},
                'y\nA\nA\n', {"project_code": "A", "region": "A"}
            ),
            (
                "A/A",
                {
                    "": {"project_code": "top", "region": "top"},
                    "A": {"project_code": "A", "region": "A"},
                },
                'y\nA/A\nA/A\n', {"project_code": "A/A", "region": "A/A"}
            ),
            (
                "A/A",
                {
                    "": {"project_code": "top", "region": "top"},
                    "A": {"project_code": "A", "region": "A"},
                },
                'y\nA\nA\n', {}
            )
        ]
    )
    def test_create_new_stack_group_folder(
        self, stack_group, config_structure, stdin, result
    ):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            os.makedirs(config_dir)

            stack_group_dir = os.path.join(project_path, "config", stack_group)
            for stack_group_path, config in config_structure.items():
                path = os.path.join(config_dir, stack_group_path)
                try:
                    os.makedirs(path)
                except OSError as e:
                    if e.errno == errno.EEXIST and os.path.isdir(path):
                        pass
                    else:
                        raise

                filepath = os.path.join(path, "config.yaml")
                with open(filepath, 'w') as config_file:
                    yaml.safe_dump(
                        config, stream=config_file, default_flow_style=False
                    )

            os.chdir(project_path)

            cmd_result = self.runner.invoke(
                cli, ["new", "group", stack_group],
                input=stdin
            )

            if result:
                with open(os.path.join(stack_group_dir, "config.yaml"))\
                        as config_file:
                    config = yaml.safe_load(config_file)
                assert config == result
            else:
                assert cmd_result.output.endswith(
                    "No config.yaml file needed - covered by parent config.\n"
                )

    def test_new_stack_group_folder_with_existing_folder(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            stack_group_dir = os.path.join(config_dir, "A")

            os.makedirs(stack_group_dir)
            os.chdir(project_path)

            cmd_result = self.runner.invoke(
                cli, ["new", "group", "A"], input="y\n\n\n"
            )

            assert cmd_result.output.startswith(
                "StackGroup path exists. "
                "Do you want initialise config.yaml?"
            )
            with open(os.path.join(
                    stack_group_dir, "config.yaml")) as config_file:
                config = yaml.safe_load(config_file)
            assert config == {"project_code": "", "region": ""}

    def test_new_stack_group_folder_with_another_exception(self):
        with self.runner.isolated_filesystem():
            project_path = os.path.abspath('./example')
            config_dir = os.path.join(project_path, "config")
            stack_group_dir = os.path.join(config_dir, "A")

            os.makedirs(stack_group_dir)
            os.chdir(project_path)
            patcher_mkdir = patch("sceptre.cli.new.os.mkdir")
            mock_mkdir = patcher_mkdir.start()
            mock_mkdir.side_effect = OSError(errno.EINVAL)
            result = self.runner.invoke(cli, ["new", "group", "A"])
            mock_mkdir = patcher_mkdir.stop()
            assert str(result.exception) == str(OSError(errno.EINVAL))

    @pytest.mark.parametrize(
        "cli_module,command,output_format,no_colour", [
            (
                'describe',
                ['describe', 'change-set', 'somepath', 'cs1'],
                'yaml',
                True
            ),
            (
                'describe',
                ['describe', 'change-set', 'somepath', 'cs1'],
                'json',
                False
            ),
            (
                'describe',
                ['describe', 'policy', 'somepolicy'],
                'yaml',
                True
            ),
            (
                'describe',
                ['describe', 'policy', 'somepolicy'],
                'json',
                False
            )
        ]
    )
    def test_write_output_format_flags(
        self, cli_module, command, output_format, no_colour
    ):
        no_colour_flag = ['--no-colour'] if no_colour else []
        output_format_flag = ['--output', output_format]
        args = output_format_flag + no_colour_flag + command

        with patch("sceptre.cli." + cli_module + ".write") as mock_write:
            self.runner.invoke(cli, args)
            mock_write.assert_called()
            for call in mock_write.call_args_list:
                args, _ = call
                assert args[1] == output_format
                assert args[2] == no_colour

    def test_setup_logging_with_debug(self):
        logger = setup_logging(True, False)
        assert logger.getEffectiveLevel() == logging.DEBUG
        assert logging.getLogger("botocore").getEffectiveLevel() == \
            logging.INFO

        # Silence logging for the rest of the tests
        logger.setLevel(logging.CRITICAL)

    def test_setup_logging_without_debug(self):
        logger = setup_logging(False, False)
        assert logger.getEffectiveLevel() == logging.INFO
        assert logging.getLogger("botocore").getEffectiveLevel() == \
            logging.CRITICAL

        # Silence logging for the rest of the tests
        logger.setLevel(logging.CRITICAL)

    @patch("sceptre.cli.click.echo")
    @pytest.mark.parametrize(
        "output_format,no_colour,expected_output", [
            ("json", True, '{\n    "stack": "CREATE_COMPLETE"\n}'),
            ("json", False, '{\n    "stack": "\x1b[32mCREATE_COMPLETE\x1b[0m\"\n}'),
            ("yaml", True, '---\nstack: CREATE_COMPLETE\n'),
            ("yaml", False, '---\nstack: \x1b[32mCREATE_COMPLETE\x1b[0m\n')
        ]
    )
    def test_write_formats(
        self, mock_echo, output_format, no_colour, expected_output
    ):
        write({"stack": "CREATE_COMPLETE"}, output_format, no_colour)
        mock_echo.assert_called_once_with(expected_output)

    @patch("sceptre.cli.click.echo")
    def test_write_status_with_colour(self, mock_echo):
        write("stack: CREATE_COMPLETE", no_colour=False)
        mock_echo.assert_called_once_with(
            '{\n    "stack": "\x1b[32mCREATE_COMPLETE\x1b[0m"\n}'
        )

    @patch("sceptre.cli.click.echo")
    def test_write_status_without_colour(self, mock_echo):
        write("stack: CREATE_COMPLETE", no_colour=True)
        mock_echo.assert_called_once_with('{\n    "stack": "CREATE_COMPLETE"\n}')

    @patch("sceptre.cli.helpers.StackStatusColourer.colour")
    @patch("sceptre.cli.helpers.logging.Formatter.format")
    def test_ColouredFormatter_format_with_string(
            self, mock_format, mock_colour
    ):
        mock_format.return_value = sentinel.response
        mock_colour.return_value = sentinel.coloured_response
        coloured_formatter = ColouredFormatter()
        response = coloured_formatter.format("string")
        mock_format.assert_called_once_with("string")
        mock_colour.assert_called_once_with(sentinel.response)
        assert response == sentinel.coloured_response

    def test_CustomJsonEncoder_with_non_json_serialisable_object(self):
        encoder = CustomJsonEncoder()
        response = encoder.encode(datetime.datetime(2016, 5, 3))
        assert response == '"2016-05-03 00:00:00"'

    def test_diff_command__diff_type_is_deepdiff__passes_deepdiff_stack_differ_to_actions(self):
        self.runner.invoke(cli, 'diff -t deepdiff dev/vpc.yaml')
        differ_used = self.mock_stack_actions.diff.call_args[0][0]
        assert isinstance(differ_used, DeepDiffStackDiffer)

    def test_diff_command__diff_type_is_difflib__passes_difflib_stack_differ_to_actions(self):
        self.runner.invoke(cli, 'diff -t difflib dev/vpc.yaml')
        differ_used = self.mock_stack_actions.diff.call_args[0][0]
        assert isinstance(differ_used, DifflibStackDiffer)

        self.runner.invoke(cli, 'diff stacks', catch_exceptions=False)

    def test_diff_command__stack_diffs_have_differences__returns_0(self):
        stacks = {deepcopy(self.mock_stack) for _ in range(3)}
        stack_name_iterator = iter(['first', 'second', 'third'])

        def fake_diff(differ):
            name = next(stack_name_iterator)
            return StackDiff(
                stack_name=name,
                template_diff=DeepDiff("I'm", "different"),
                config_diff=DeepDiff("same", "same"),
                is_deployed=True,
                generated_config=None,
                generated_template=None
            )

        self.mock_stack_actions.diff.side_effect = fake_diff
        self.mock_config_reader.construct_stacks.return_value = (stacks, stacks)

        result = self.runner.invoke(cli, 'diff stacks', catch_exceptions=False)
        assert result.exit_code == 0

    def test_diff_command__no_differences__returns_0(self):
        stacks = {deepcopy(self.mock_stack) for _ in range(3)}
        stack_name_iterator = iter(['first', 'second', 'third'])

        def fake_diff(differ):
            name = next(stack_name_iterator)
            return StackDiff(
                stack_name=name,
                template_diff=DeepDiff("same", "same"),
                config_diff=DeepDiff("same", "same"),
                is_deployed=True,
                generated_config=None,
                generated_template=None
            )

        self.mock_stack_actions.diff.side_effect = fake_diff
        self.mock_config_reader.construct_stacks.return_value = (stacks, stacks)

        result = self.runner.invoke(cli, 'diff stacks', catch_exceptions=False)
        assert result.exit_code == 0

    @pytest.mark.parametrize(
        ['bar'],
        [('**********',), ('----------',)]
    )
    def test_diff_command__bars_are_all_full_width_of_output(self, bar):
        stacks = {deepcopy(self.mock_stack) for _ in range(3)}
        stack_name_iterator = iter(['first', 'second', 'third'])

        def fake_diff(differ):
            name = next(stack_name_iterator)
            return StackDiff(
                stack_name=name,
                template_diff=DeepDiff("same", "same"),
                config_diff=DeepDiff("same", "same"),
                is_deployed=True,
                generated_config=None,
                generated_template=None
            )

        self.mock_stack_actions.diff.side_effect = fake_diff
        self.mock_config_reader.construct_stacks.return_value = (stacks, stacks)

        result = self.runner.invoke(cli, 'diff stacks', catch_exceptions=False)
        output_lines = result.stdout.splitlines()
        max_line_length = len(max(output_lines, key=len))
        star_bars = [line for line in output_lines if bar in line]
        assert all(len(line) == max_line_length for line in star_bars)

    @pytest.mark.parametrize("input,expected_output", [
        (
            {"a_dict": '{"with_embedded":"json"}'},
            {"a_dict": {"with_embedded": "json"}}
        ),
        (
            {"a_dict": ['{"with_embedded":"json"}']},
            {"a_dict": [{"with_embedded": "json"}]}
        ),
    ])
    def test_deserialize_json_properties(self, input, expected_output):
        output = deserialize_json_properties(input)
        assert output == expected_output

    def test_drift_detect(self):
        self.mock_stack_actions.drift_detect.return_value = {
            "StackId": "fake-stack-id",
            "StackDriftDetectionId": "3fb76910-f660-11eb-80ac-0246f7a6da62",
            "StackDriftStatus": "IN_SYNC",
            "DetectionStatus": "DETECTION_COMPLETE",
            "DriftedStackResourceCount": 0
        }
        result = self.runner.invoke(
            cli, ["drift", "detect", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert result.output == (
            '---\n'
            'mock-stack-external:\n'
            '  DetectionStatus: DETECTION_COMPLETE\n'
            '  DriftedStackResourceCount: 0\n'
            '  StackDriftDetectionId: 3fb76910-f660-11eb-80ac-0246f7a6da62\n'
            '  StackDriftStatus: IN_SYNC\n'
            '  StackId: fake-stack-id\n\n'
        )

    def test_drift_show(self):
        self.mock_stack_actions.drift_show.return_value = (
            "DETECTION_COMPLETE", {"some": "json"}
        )
        result = self.runner.invoke(
            cli, ["drift", "show", "dev/vpc.yaml"]
        )
        assert result.exit_code == 0
        assert result.output == "---\nmock-stack-external:\n  some: json\n\n"



# -*- coding: utf-8 -*-
import json

import pytest

from sceptre.exceptions import UnsupportedTemplateFileTypeError
from sceptre.template_handlers.http import Http
from unittest.mock import patch


class TestHttp(object):

    def test_get_template(self, requests_mock):
        url = "https://raw.githubusercontent.com/acme/bucket.yaml"
        requests_mock.get(url, content=b"Stuff is working")
        template_handler = Http(
            name="vpc",
            arguments={"url": url},
        )
        result = template_handler.handle()
        assert result == b"Stuff is working"

    def test_handler_unsupported_type(self):
        handler = Http("http_handler", {'url': 'https://raw.githubusercontent.com/acme/bucket.unsupported'})
        with pytest.raises(UnsupportedTemplateFileTypeError):
            handler.handle()

    @pytest.mark.parametrize("url", [
        ("https://raw.githubusercontent.com/acme/bucket.json"),
        ("https://raw.githubusercontent.com/acme/bucket.yaml"),
        ("https://raw.githubusercontent.com/acme/bucket.template")
    ])
    @patch('sceptre.template_handlers.http.Http._get_template')
    def test_handler_raw_template(self, mock_get_template, url):
        mock_get_template.return_value = {}
        handler = Http("http_handler", {'url': url})
        handler.handle()
        assert mock_get_template.call_count == 1

    @patch('sceptre.template_handlers.helper.render_jinja_template')
    @patch('sceptre.template_handlers.http.Http._get_template')
    def test_handler_jinja_template(self, mock_get_template, mock_render_jinja_template):
        mock_get_template_response = {
            "Description": "test template",
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {
                "touchNothing": {
                    "Type": "AWS::CloudFormation::WaitConditionHandle"
                }
            }
        }
        mock_get_template.return_value = json.dumps(mock_get_template_response).encode('utf-8')
        handler = Http("http_handler", {'url': 'https://raw.githubusercontent.com/acme/bucket.j2'})
        handler.handle()
        assert mock_render_jinja_template.call_count == 1

    @patch('sceptre.template_handlers.helper.call_sceptre_handler')
    @patch('sceptre.template_handlers.http.Http._get_template')
    def test_handler_python_template(self, mock_get_template, mock_call_sceptre_handler):
        mock_get_template_response = {
            "Description": "test template",
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {
                "touchNothing": {
                    "Type": "AWS::CloudFormation::WaitConditionHandle"
                }
            }
        }
        mock_get_template.return_value = json.dumps(mock_get_template_response).encode('utf-8')
        handler = Http("http_handler", {'url': 'https://raw.githubusercontent.com/acme/bucket.py'})
        handler.handle()
        assert mock_call_sceptre_handler.call_count == 1

    @patch('sceptre.template_handlers.helper.call_sceptre_handler')
    @patch('sceptre.template_handlers.http.Http._get_template')
    def test_handler_override_handler_options(self, mock_get_template, mock_call_sceptre_handler):
        mock_get_template_response = {
            "Description": "test template",
            "AWSTemplateFormatVersion": "2010-09-09",
            "Resources": {
                "touchNothing": {
                    "Type": "AWS::CloudFormation::WaitConditionHandle"
                }
            }
        }
        mock_get_template.return_value = json.dumps(mock_get_template_response).encode('utf-8')
        custom_handler_options = {"timeout": 10, "retries": 20}
        handler = Http("http_handler",
                       {'url': 'https://raw.githubusercontent.com/acme/bucket.py'},
                       stack_group_config={"http_template_handler": custom_handler_options}
                       )
        handler.handle()
        assert mock_get_template.call_count == 1
        args, options = mock_get_template.call_args
        assert options == {"timeout": 10, "retries": 20}



# -*- coding: utf-8 -*-

"""
sceptre.connection_manager

This module implements a ConnectionManager class, which simplifies and manages
Boto3 calls.
"""

import functools
import logging
import random
import threading
import time
import boto3

from os import environ
from botocore.exceptions import ClientError

from sceptre.helpers import mask_key
from sceptre.exceptions import InvalidAWSCredentialsError, RetryLimitExceededError


def _retry_boto_call(func):
    """
    Retries a Boto3 call up to 30 times if request rate limits are hit.

    Between each try we wait a random amount with a max of that time being 45 seconds.
    Specifically we are picking number between a ceiling (delay_cap) of 45 seconds and the last
    delay multiplied by 2.5, rounding to two decimal places.  You can read more
    here: https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    If rate limits are hit 30 times, _retry_boto_call raises a
    sceptre.exceptions.RetryLimitExceededException.

    :param func: A function that uses boto calls
    :type func: function
    :returns: The decorated function.
    :rtype: function
    :raises: sceptre.exceptions.RetryLimitExceededException
    """
    logger = logging.getLogger(__name__)

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        max_retries = 30
        attempts = 1
        mdelay = 1
        delay_cap = 45
        while attempts < max_retries:
            try:
                return func(*args, **kwargs)
            except ClientError as e:
                if e.response["Error"]["Code"] == "Throttling":
                    logger.error("Request limit exceeded, pausing {}...".format(mdelay))
                    time.sleep(mdelay)

                    # Using De-correlated Jitter Algorithm
                    # We are picking number between a ceiling (delay_cap) of 45 seconds and the
                    # last delay multiplied by 2.5, rounding to two decimal places.
                    mdelay = min(delay_cap, round((random.uniform(1, mdelay * 2.5)), 2))

                    attempts += 1
                else:
                    raise
        raise RetryLimitExceededError(
            "Exceeded request limit {0} times. Aborting.".format(
                max_retries
            )
        )

    return decorated


class ConnectionManager(object):
    """
    The Connection Manager is used to create boto3 clients for
    the various AWS services that Sceptre needs to interact with.

    :param profile: The AWS credentials profile that should be used.
    :type profile: str
    :param iam_role: The iam_role that should be assumed in the account.
    :type iam_role: str
    :param stack_name: The CloudFormation stack name for this connection.
    :type stack_name: str
    :param region: The region to use.
    :type region: str
    """

    _session_lock = threading.Lock()
    _client_lock = threading.Lock()
    _boto_sessions = {}
    _clients = {}
    _stack_keys = {}

    def __init__(
        self, region, profile=None, stack_name=None,
        iam_role=None, iam_role_session_duration=None
    ):

        self.logger = logging.getLogger(__name__)

        self.region = region
        self.profile = profile
        self.stack_name = stack_name
        self.iam_role = iam_role
        self.iam_role_session_duration = iam_role_session_duration

        if stack_name:
            self._stack_keys[stack_name] = (region, profile, iam_role)

    def __repr__(self):
        return (
            "sceptre.connection_manager.ConnectionManager(region='{0}', "
            "profile='{1}', stack_name='{2}', iam_role='{3}', iam_role_session_duration='{4}')".format(
                self.region, self.profile, self.stack_name, self.iam_role, self.iam_role_session_duration
            )
        )

    def _get_session(self, profile, region, iam_role):
        """
        Returns a boto session in the target account.

        If a ``profile`` is specified in ConnectionManager's initialiser,
        then the profile is used to generate temporary credentials to create
        the Boto session. If ``profile`` is not specified then the default
        profile is assumed to create the boto session.

        :returns: The Boto3 session.
        :rtype: boto3.session.Session
        :raises: botocore.exceptions.ClientError
        """
        with self._session_lock:
            self.logger.debug("Getting Boto3 session")
            key = (region, profile, iam_role)

            if self._boto_sessions.get(key) is None:
                self.logger.debug("No Boto3 session found, creating one...")
                self.logger.debug("Using cli credentials...")

                # Credentials from env take priority over profile
                config = {
                    "profile_name": profile,
                    "region_name": region,
                    "aws_access_key_id": environ.get("AWS_ACCESS_KEY_ID"),
                    "aws_secret_access_key": environ.get("AWS_SECRET_ACCESS_KEY"),
                    "aws_session_token": environ.get("AWS_SESSION_TOKEN")
                }

                session = boto3.session.Session(**config)
                self._boto_sessions[key] = session

                if session.get_credentials() is None:
                    raise InvalidAWSCredentialsError(
                        "Session credentials were not found. Profile: {0}. Region: {1}.".format(
                            config["profile_name"], config["region_name"]
                        )
                    )

                if iam_role:
                    sts_client = session.client("sts")
                    # maximum session name length is 64 chars. 56 + "-session" = 64
                    session_name = f'{iam_role.split("/")[-1][:56]}-session'
                    assume_role_kwargs = {
                        'RoleArn': iam_role,
                        'RoleSessionName': session_name,
                    }
                    if self.iam_role_session_duration:
                        assume_role_kwargs['DurationSeconds'] = self.iam_role_session_duration
                    sts_response = sts_client.assume_role(**assume_role_kwargs)

                    credentials = sts_response["Credentials"]
                    session = boto3.session.Session(
                        aws_access_key_id=credentials["AccessKeyId"],
                        aws_secret_access_key=credentials["SecretAccessKey"],
                        aws_session_token=credentials["SessionToken"],
                        region_name=region
                    )

                    if session.get_credentials() is None:
                        raise InvalidAWSCredentialsError(
                            "Session credentials were not found. Role: {0}. Region: {1}.".format(
                                iam_role, region
                            )
                        )

                    self._boto_sessions[key] = session

                self.logger.debug(
                    "Using credential set from %s: %s",
                    session.get_credentials().method,
                    {
                        "AccessKeyId": mask_key(
                            session.get_credentials().access_key
                        ),
                        "SecretAccessKey": mask_key(
                            session.get_credentials().secret_key
                        ),
                        "Region": session.region_name
                    }
                )

                self.logger.debug("Boto3 session created")

            return self._boto_sessions[key]

    def _get_client(self, service, region, profile, stack_name, iam_role):
        """
        Returns the Boto3 client associated with <service>.

        Equivalent to calling Boto3.client(<service>). Gets the client using
        ``boto_session``.

        :param service: The Boto3 service to return a client for.
        :type service: str
        :returns: The Boto3 client.
        :rtype: boto3.client.Client
        """
        with self._client_lock:
            key = (service, region, profile, stack_name, iam_role)
            if self._clients.get(key) is None:
                self.logger.debug(
                    "No %s client found, creating one...", service
                )
                self._clients[key] = self._get_session(
                    profile, region, iam_role
                ).client(service)

            return self._clients[key]

    @_retry_boto_call
    def call(
        self, service, command, kwargs=None, profile=None, region=None,
        stack_name=None, iam_role=None
    ):
        """
        Makes a thread-safe Boto3 client call.

        Equivalent to ``boto3.client(<service>).<command>(**kwargs)``.

        :param service: The Boto3 service to return a client for.
        :type service: str
        :param command: The Boto3 command to call.
        :type command: str
        :param kwargs: The keyword arguments to supply to <command>.
        :type kwargs: dict
        :returns: The response from the Boto3 call.
        :rtype: dict
        """
        if region is None and profile is None and iam_role is None:
            if stack_name and stack_name in self._stack_keys:
                region, profile, iam_role = self._stack_keys[stack_name]
            else:
                region = self.region
                profile = self.profile
                iam_role = self.iam_role

        if kwargs is None:  # pragma: no cover
            kwargs = {}

        client = self._get_client(service, region, profile, stack_name, iam_role)
        return getattr(client, command)(**kwargs)



from typing import Any, List

from sceptre.resolvers import Resolver


class StackAttr(Resolver):
    """Resolves to the value of another field on the Stack Config, including other resolvers.

    The argument for this resolver should be the "key path" from the stack object, which can access
    nested keys/indexes using a "." to separate segments.

    For example, given this Stack Config structure...

        sceptre_user_data:
            nested_list:
              - first
              - second

    Using "!stack_attr sceptre_user_data.nested_list.1" on your stack would resolve to "second".
    """

    # These are all the attributes on Stack Configs whose names are changed when they are assigned
    # to the Stack instance.
    STACK_ATTR_MAP = {
        'template': 'template_handler_config',
        'protect': 'protected',
        'stack_name': 'external_name',
        'stack_tags': 'tags'
    }

    def resolve(self) -> Any:
        """Returns the resolved value of the field referenced by the resolver's argument."""
        segments = self.argument.split('.')

        # Remap top-level attributes to match stack config
        first_segment = segments[0]
        segments[0] = self.STACK_ATTR_MAP.get(first_segment, first_segment)

        if self._key_is_from_stack_group_config(first_segment):
            obj = self.stack.stack_group_config
        else:
            obj = self.stack

        result = self._recursively_resolve_segments(obj, segments)
        return result

    def _key_is_from_stack_group_config(self, key: str):
        return key in self.stack.stack_group_config and not hasattr(self.stack, key)

    def _recursively_resolve_segments(self, obj: Any, segments: List[str]):
        if not segments:
            return obj

        attr_name, *rest = segments
        if isinstance(obj, dict):
            value = obj[attr_name]
        elif isinstance(obj, list):
            value = obj[int(attr_name)]
        else:
            value = getattr(obj, attr_name)

        return self._recursively_resolve_segments(value, rest)



import functools
import itertools
from collections import defaultdict
from typing import Optional, List, Set
from unittest.mock import create_autospec, Mock

import pytest

from sceptre.cli.launch import Launcher
from sceptre.cli.prune import Pruner
from sceptre.context import SceptreContext
from sceptre.exceptions import DependencyDoesNotExistError
from sceptre.plan.plan import SceptrePlan
from sceptre.stack import Stack
from sceptre.stack_status import StackStatus


class FakePlan(SceptrePlan):
    def __init__(
        self,
        context: SceptreContext,
        command_stacks: Set[Stack],
        all_stacks: Set[Stack],
        statuses_to_return: dict,
    ):
        self.context = context
        self.command = None
        self.reverse = None
        self.launch_order: Optional[List[Set[Stack]]] = None

        self.all_stacks = all_stacks
        self.command_stacks = command_stacks
        self.statuses_to_return = statuses_to_return

        self.executions = []

    def _execute(self, *args):
        self.executions.append(
            (self.command, self.launch_order.copy(), args)
        )
        return {
            stack: self.statuses_to_return[stack]
            for stack in self
        }

    def _generate_launch_order(self, reverse=False) -> List[Set[Stack]]:
        launch_order = [self.command_stacks]
        if self.context.ignore_dependencies:
            return launch_order

        all_stacks = list(self.all_stacks)

        for start_index in range(0, len(all_stacks), 2):
            chunk = {
                stack for stack in all_stacks[start_index: start_index + 2]
                if stack not in self.command_stacks
                and self._has_dependency_on_a_command_stack(stack)
            }
            if len(chunk):
                launch_order.append(chunk)

        return launch_order

    @functools.lru_cache()
    def _has_dependency_on_a_command_stack(self, stack):
        if len(self.command_stacks.intersection(stack.dependencies)):
            return True

        for dependency in stack.dependencies:
            if self._has_dependency_on_a_command_stack(dependency):
                return True

        return False


class TestLauncher:

    def setup_method(self, test_method):
        self.plans: List[FakePlan] = []

        self.context = SceptreContext(
            project_path="project",
            command_path="my-test-group",
        )
        self.cloned_context = self.context.clone()
        # Since contexts don't have a __eq__ method, you can't assert easily off the result of
        # clone without some hijinks.
        self.context = Mock(wraps=self.context, **{
            'clone.return_value': self.cloned_context,
            'ignore_dependencies': False
        })

        self.all_stacks = [
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[]),
            Mock(spec=Stack, ignore=False, obsolete=False, dependencies=[])
        ]
        for index, stack in enumerate(self.all_stacks):
            stack.name = f'stacks/stack-{index}.yaml'

        self.command_stacks = list(self.all_stacks)

        self.statuses_to_return = defaultdict(lambda: StackStatus.COMPLETE)

        self.fake_pruner = Mock(spec=Pruner, **{
            'prune.return_value': 0
        })

        self.plan_factory = create_autospec(SceptrePlan)
        self.plan_factory.side_effect = self.fake_plan_factory
        self.pruner_factory = create_autospec(Pruner)
        self.pruner_factory.return_value = self.fake_pruner

        self.launcher = Launcher(self.context, self.plan_factory, self.pruner_factory)

    def fake_plan_factory(self, sceptre_context):
        fake_plan = FakePlan(
            sceptre_context,
            set(self.command_stacks),
            set(self.all_stacks),
            self.statuses_to_return
        )
        self.plans.append(fake_plan)
        return fake_plan

    def get_executed_stacks(self, plan_number: int):
        launch_order = self.plans[plan_number].executions[0][1]
        return list(itertools.chain.from_iterable(launch_order))

    def test_launch__launches_stacks_that_are_neither_ignored_nor_obsolete(self):
        assert all(not s.ignore and not s.obsolete for s in self.all_stacks)
        self.command_stacks = self.all_stacks
        self.launcher.launch(True)
        launched_stacks = set(self.get_executed_stacks(0))
        expected_stacks = set(self.all_stacks)
        assert expected_stacks == launched_stacks
        assert self.plans[0].executions[0][0] == "launch"
        assert len(self.plans[0].executions) == 1

    def test_launch__prune__no_obsolete_stacks__does_not_delete_any_stacks(self):
        assert all(not s.obsolete for s in self.all_stacks)
        self.launcher.launch(True)
        assert len(self.plans) == 1
        assert self.plans[0].executions[0][0] == "launch"

    def test_launch__prune__instantiates_and_invokes_pruner(self):
        self.launcher.launch(True)
        self.fake_pruner.prune.assert_any_call()

    def test_launch__no_prune__obsolete_stacks__does_not_delete_any_stacks(self):
        self.all_stacks[4].obsolete = True
        self.all_stacks[5].obsolete = True

        self.launcher.launch(False)
        assert len(self.plans) == 1
        assert self.plans[0].executions[0][0] == "launch"

    @pytest.mark.parametrize("prune", [
        pytest.param(True, id="prune"),
        pytest.param(False, id="no prune")
    ])
    def test_launch__returns_0(self, prune):
        assert all(not s.ignore and not s.obsolete for s in self.all_stacks)
        result = self.launcher.launch(prune)

        assert result == 0

    def test_launch__does_not_launch_stacks_that_should_be_excluded(self):
        self.all_stacks[4].ignore = True
        self.all_stacks[5].obsolete = True

        self.launcher.launch(True)

        launched_stacks = set(self.get_executed_stacks(0))
        expected_stacks = {s for i, s in enumerate(self.all_stacks) if i not in (4, 5)}
        assert expected_stacks == launched_stacks
        assert self.plans[0].executions[0][0] == "launch"

    def test_launch__prune__stack_with_dependency_marked_obsolete__raises_dependency_does_not_exist_error(self):
        self.all_stacks[0].obsolete = True
        self.all_stacks[1].dependencies.append(self.all_stacks[0])

        self.command_stacks = [self.all_stacks[0]]

        with pytest.raises(DependencyDoesNotExistError):
            self.launcher.launch(True)

    def test_launch__prune__ignore_dependencies__stack_with_dependency_marked_obsolete__raises_no_error(self):
        self.all_stacks[0].obsolete = True
        self.all_stacks[1].dependencies.append(self.all_stacks[0])

        self.command_stacks = [self.all_stacks[0]]
        self.context.ignore_dependencies = True
        self.launcher.launch(True)

    def test_launch__no_prune__does_not_raise_error(self):
        self.all_stacks[0].obsolete = True
        self.all_stacks[1].dependencies.append(self.all_stacks[0])
        self.launcher.launch(False)

    def test_launch__stacks_are_pruned__delete_and_deploy_actions_succeed__returns_0(self):
        self.all_stacks[0].obsolete = True

        code = self.launcher.launch(True)
        assert code == 0

    def test_launch__pruner_returns_nonzero__returns_nonzero(self):
        self.fake_pruner.prune.return_value = 99

        code = self.launcher.launch(True)
        assert code == 99

    def test_launch__deploy_action_fails__returns_nonzero(self):
        self.statuses_to_return[self.all_stacks[3]] = StackStatus.FAILED
        code = self.launcher.launch(False)
        assert code != 0



# -*- coding: utf-8 -*-

import json
import yaml
import os
import threading

import pytest
from unittest.mock import patch, sentinel, Mock

from freezegun import freeze_time
from botocore.exceptions import ClientError

from sceptre.template import Template
from sceptre.connection_manager import ConnectionManager
from sceptre.exceptions import UnsupportedTemplateFileTypeError
from sceptre.exceptions import TemplateSceptreHandlerError, TemplateNotFoundError
from sceptre.template_handlers import TemplateHandler


class MockTemplateHandler(TemplateHandler):
    def __init__(self, *args, **kwargs):
        super(MockTemplateHandler, self).__init__(*args, **kwargs)

    def schema(self):
        return {}

    def handle(self):
        return self.arguments["argument"]


class TestTemplate(object):

    def setup_method(self, test_method):
        self.region = "region"
        self.bucket_name = "bucket_name"
        self.stack_group_path = "stack_group_path"
        self.stack_name = "stack_name"

        connection_manager = Mock(spec=ConnectionManager)
        connection_manager.create_bucket_lock = threading.Lock()

        self.template = Template(
            name="template_name",
            handler_config={"type": "file", "path": "/folder/template.py"},
            sceptre_user_data={},
            stack_group_config={
                "project_path": "projects"
            },
            connection_manager=connection_manager,
        )

    def test_initialise_template_default_handler_type(self):
        template = Template(
            name="template_name",
            handler_config={"path": "/folder/template.py"},
            sceptre_user_data={},
            stack_group_config={},
            connection_manager={},
        )

        assert template.handler_config == {"type": "file", "path": "/folder/template.py"}

    def test_initialise_template(self):
        assert self.template.handler_config == {"type": "file", "path": "/folder/template.py"}
        assert self.template.name == "template_name"
        assert self.template.sceptre_user_data == {}
        assert self.template._body is None

    def test_repr(self):
        representation = self.template.__repr__()
        assert representation == "sceptre.template.Template(" \
            "name='template_name', handler_config={'type': 'file', 'path': '/folder/template.py'}" \
            ", sceptre_user_data={}, s3_details=None)"

    def test_body_with_cache(self):
        self.template._body = sentinel.body
        body = self.template.body
        assert body == sentinel.body

    @freeze_time("2012-01-01")
    @patch("sceptre.template.Template._bucket_exists")
    def test_upload_to_s3_with_valid_s3_details(self, mock_bucket_exists):
        self.template._body = '{"template": "mock"}'
        mock_bucket_exists.return_value = True
        self.template.s3_details = {
            "bucket_name": "bucket-name",
            "bucket_key": "bucket-key"
        }

        self.template.upload_to_s3()

        get_bucket_location_call, put_object_call = self.template.connection_manager.call.call_args_list
        get_bucket_location_call.assert_called_once_with(
            service="s3",
            command="get_bucket_location",
            kwargs={
                "Bucket": "bucket-name"
            }
        )
        put_object_call.assert_called_once_with(
            service="s3",
            command="put_object",
            kwargs={
                "Bucket": "bucket-name",
                "Key": "bucket-key",
                "Body": '{"template": "mock"}',
                "ServerSideEncryption": "AES256"
            }
        )

    def test_domain_from_region(self):
        assert self.template._domain_from_region("us-east-1") == "com"
        assert self.template._domain_from_region("cn-north-1") == "com.cn"
        assert self.template._domain_from_region("cn-northwest-1") == "com.cn"

    def test_bucket_exists_with_bucket_that_exists(self):
        # connection_manager.call doesn't raise an exception, mimicing the
        # behaviour when head_bucket successfully executes.
        self.template.s3_details = {
            "bucket_name": "bucket-name",
            "bucket_key": "bucket-key"
        }

        assert self.template._bucket_exists() is True

    def test_create_bucket_with_unreadable_bucket(self):
        self.template.connection_manager.region = "eu-west-1"
        self.template.s3_details = {
            "bucket_name": "bucket-name",
            "bucket_key": "bucket-key"
        }

        self.template.connection_manager.call.side_effect = ClientError(
            {
                "Error": {
                    "Code": 500,
                    "Message": "Bucket Unreadable"
                }
            },
            sentinel.operation
        )
        with pytest.raises(ClientError) as e:
            self.template._create_bucket()
            assert e.value.response["Error"]["Code"] == 500
            assert e.value.response["Error"]["Message"] == "Bucket Unreadable"

    def test_bucket_exists_with_non_existent_bucket(self):
        # connection_manager.call is called twice, and should throw the
        # Not Found ClientError only for the first call.
        self.template.s3_details = {
            "bucket_name": "bucket-name",
            "bucket_key": "bucket-key"
        }

        self.template.connection_manager.call.side_effect = [
            ClientError(
                {
                    "Error": {
                        "Code": 404,
                        "Message": "Not Found"
                    }
                },
                sentinel.operation
            ),
            None
        ]

        existance = self.template._bucket_exists()

        assert existance is False

    def test_create_bucket_in_us_east_1(self):
        # connection_manager.call is called twice, and should throw the
        # Not Found ClientError only for the first call.
        self.template.connection_manager.region = "us-east-1"
        self.template.s3_details = {
            "bucket_name": "bucket-name",
            "bucket_key": "bucket-key"
        }

        self.template._create_bucket()

        self.template.connection_manager.call.assert_any_call(
            service="s3",
            command="create_bucket",
            kwargs={"Bucket": "bucket-name"}
        )

    @patch("sceptre.template.Template.upload_to_s3")
    def test_get_boto_call_parameter_with_s3_details(self, mock_upload_to_s3):
        # self.stack._template = Mock(spec=Template)
        mock_upload_to_s3.return_value = sentinel.template_url
        self.template.s3_details = {
            "bucket_name": sentinel.bucket_name,
            "bucket_key": sentinel.bucket_key
        }

        boto_parameter = self.template.get_boto_call_parameter()

        assert boto_parameter == {"TemplateURL": sentinel.template_url}

    def test_get_boto_call_parameter__has_s3_details_but_bucket_name_is_none__gets_template_body_dict(self):
        self.template._body = sentinel.body
        self.template.s3_details = {
            "bucket_name": None,
            "bucket_key": sentinel.bucket_key
        }

        boto_parameter = self.template.get_boto_call_parameter()

        assert boto_parameter == {"TemplateBody": sentinel.body}

    def test_get_template_details_without_upload(self):
        self.template.s3_details = None
        self.template._body = sentinel.body
        boto_parameter = self.template.get_boto_call_parameter()

        assert boto_parameter == {"TemplateBody": sentinel.body}

    def test_body_with_json_template(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures-vpc/templates/vpc.json"
        )
        output = self.template.body
        output_dict = yaml.safe_load(output)
        with open("tests/fixtures/templates/compiled_vpc.json", "r") as f:
            expected_output_dict = json.loads(f.read())
        assert output_dict == expected_output_dict

    def test_body_with_yaml_template(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.yaml"
        )
        output = self.template.body
        output_dict = yaml.safe_load(output)
        with open("tests/fixtures/templates/compiled_vpc.json", "r") as f:
            expected_output_dict = json.loads(f.read())
        assert output_dict == expected_output_dict

    def test_body_with_generic_template(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.template"
        )
        output = self.template.body
        output_dict = yaml.safe_load(output)
        with open("tests/fixtures/templates/compiled_vpc.json", "r") as f:
            expected_output_dict = json.loads(f.read())
        assert output_dict == expected_output_dict

    def test_body_with_chdir_template(self):
        self.template.sceptre_user_data = None
        self.template.name = "chdir"
        current_dir = os.getcwd()
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/chdir.py"
        )
        try:
            yaml.safe_load(self.template.body)
        except ValueError:
            assert False
        finally:
            os.chdir(current_dir)

    def test_body_with_missing_file(self):
        self.template.handler_config["path"] = "incorrect/template/path.py"
        with pytest.raises(TemplateNotFoundError):
            self.template.body

    def test_body_with_python_template(self):
        self.template.sceptre_user_data = None
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.py"
        )
        actual_output = yaml.safe_load(self.template.body)
        with open("tests/fixtures/templates/compiled_vpc.json", "r") as f:
            expected_output = json.loads(f.read())
        assert actual_output == expected_output

    def test_body_with_python_template_with_sgt(self):
        self.template.sceptre_user_data = None
        self.template.name = "vpc_sgt"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc_sgt.py"
        )
        actual_output = yaml.safe_load(self.template.body)
        with open("tests/fixtures/templates/compiled_vpc.json", "r") as f:
            expected_output = json.loads(f.read())
        assert actual_output == expected_output

    def test_body_injects_yaml_start_marker(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.without_start_marker.yaml"
        )
        output = self.template.body
        with open("tests/fixtures/templates/vpc.yaml", "r") as f:
            expected_output = f.read()
        assert output == expected_output

    def test_body_with_existing_yaml_start_marker(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.yaml"
        )
        output = self.template.body
        with open("tests/fixtures/templates/vpc.yaml", "r") as f:
            expected_output = f.read()
        assert output == expected_output

    def test_body_with_existing_yaml_start_marker_j2(self):
        self.template.name = "vpc"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc.yaml.j2"
        )
        self.template.sceptre_user_data = {
            "vpc_id": "10.0.0.0/16"
        }
        output = self.template.body
        with open("tests/fixtures/templates/compiled_vpc.yaml", "r") as f:
            expected_output = f.read()
        assert output == expected_output.rstrip()

    def test_body_injects_sceptre_user_data(self):
        self.template.sceptre_user_data = {
            "cidr_block": "10.0.0.0/16"
        }
        self.template.name = "vpc_sud"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc_sud.py"
        )

        actual_output = yaml.safe_load(self.template.body)
        with open("tests/fixtures/templates/compiled_vpc_sud.json", "r") as f:
            expected_output = json.loads(f.read())
        assert actual_output == expected_output

    def test_body_injects_sceptre_user_data_incorrect_function(self):
        self.template.sceptre_user_data = {
            "cidr_block": "10.0.0.0/16"
        }
        self.template.name = "vpc_sud_incorrect_function"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc_sud_incorrect_function.py"
        )
        with pytest.raises(TemplateSceptreHandlerError):
            self.template.body

    def test_body_injects_sceptre_user_data_incorrect_handler(self):
        self.template.sceptre_user_data = {
            "cidr_block": "10.0.0.0/16"
        }
        self.template.name = "vpc_sud_incorrect_handler"
        self.template.handler_config["path"] = os.path.join(
            os.getcwd(),
            "tests/fixtures/templates/vpc_sud_incorrect_handler.py"
        )
        with pytest.raises(TypeError):
            self.template.body

    def test_body_with_incorrect_filetype(self):
        self.template.handler_config["path"] = (
            "path/to/something.ext"
        )
        with pytest.raises(UnsupportedTemplateFileTypeError):
            self.template.body

    def test_template_handler_is_called(self):
        self.template.handler_config = {
            "type": "test",
            "argument": sentinel.template_handler_argument
        }

        self.template._registry = {
            "test": MockTemplateHandler
        }

        result = self.template.body
        assert result == "---\n" + str(sentinel.template_handler_argument)



# -*- coding: utf-8 -*-

from colorama import init, Fore, Style
from sceptre.stack_status_colourer import StackStatusColourer


class TestStackStatusColourer(object):

    def setup_method(self, test_method):
        init()
        self.stack_status_colourer = StackStatusColourer()
        self.statuses = {
            "CREATE_COMPLETE": Fore.GREEN,
            "CREATE_FAILED": Fore.RED,
            "CREATE_IN_PROGRESS": Fore.YELLOW,
            "DELETE_COMPLETE": Fore.GREEN,
            "DELETE_FAILED": Fore.RED,
            "DELETE_IN_PROGRESS": Fore.YELLOW,
            "ROLLBACK_COMPLETE": Fore.RED,
            "ROLLBACK_FAILED": Fore.RED,
            "ROLLBACK_IN_PROGRESS": Fore.YELLOW,
            "UPDATE_COMPLETE": Fore.GREEN,
            "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS": Fore.YELLOW,
            "UPDATE_FAILED": Fore.RED,
            "UPDATE_IN_PROGRESS": Fore.YELLOW,
            "UPDATE_ROLLBACK_COMPLETE": Fore.GREEN,
            "UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS": Fore.YELLOW,
            "UPDATE_ROLLBACK_FAILED": Fore.RED,
            "UPDATE_ROLLBACK_IN_PROGRESS": Fore.YELLOW
        }

    def test_colour_with_string_with_no_stack_statuses(self):
        response = self.stack_status_colourer.colour("string with no statuses")
        assert response == "string with no statuses"

    def test_colour_with_string_with_single_stack_status(self):
        strings = [
            "string string {0} string".format(status)
            for status in sorted(self.statuses.keys())
        ]

        responses = [
            self.stack_status_colourer.colour(string)
            for string in strings
        ]

        assert responses == [
            "string string {0}{1}{2} string".format(
                self.statuses[status],
                status,
                Style.RESET_ALL
            )
            for status in sorted(self.statuses.keys())
        ]

    def test_colour_with_string_with_multiple_stack_statuses(self):
        response = self.stack_status_colourer.colour(
            " ".join(sorted(self.statuses.keys()))
        )
        assert response == " ".join([
            "{0}{1}{2}".format(
                self.statuses[status],
                status,
                Style.RESET_ALL
            )
            for status in sorted(self.statuses.keys())
        ])



# -*- coding: utf-8 -*-

from troposphere import Template, Parameter, Ref, Output

from troposphere.ec2 import VPC, InternetGateway, VPCGatewayAttachment

t = Template()

cidr_block_param = t.add_parameter(Parameter(
    "CidrBlock",
    Type="String",
    Default="10.0.0.0/16",
))

vpc = t.add_resource(VPC(
    "VirtualPrivateCloud",
    CidrBlock=Ref(cidr_block_param),
    InstanceTenancy="default",
    EnableDnsSupport=True,
    EnableDnsHostnames=True,
))

igw = t.add_resource(InternetGateway(
    "InternetGateway",
))

igw_attachment = t.add_resource(VPCGatewayAttachment(
    "IGWAttachment",
    VpcId=Ref(vpc),
    InternetGatewayId=Ref(igw),
))

vpc_id_output = t.add_output(Output(
    "VpcId",
    Description="New VPC ID",
    Value=Ref(vpc)
))



import pytest
from unittest.mock import MagicMock, patch, sentinel

from sceptre.context import SceptreContext
from sceptre.stack import Stack
from sceptre.config.reader import ConfigReader
from sceptre.plan.plan import SceptrePlan


class TestSceptrePlan(object):

    def setup_method(self, test_method):
        self.patcher_SceptrePlan = patch("sceptre.plan.plan.SceptrePlan")
        self.stack = Stack(
            name='dev/app/stack', project_code=sentinel.project_code,
            template_path=sentinel.template_path, region=sentinel.region,
            profile=sentinel.profile, parameters={"key1": "val1"},
            sceptre_user_data=sentinel.sceptre_user_data, hooks={},
            s3_details=None, dependencies=sentinel.dependencies,
            role_arn=sentinel.role_arn, protected=False,
            tags={"tag1": "val1"}, external_name=sentinel.external_name,
            notifications=[sentinel.notification],
            on_failure=sentinel.on_failure,
            stack_timeout=sentinel.stack_timeout
        )
        self.mock_context = MagicMock(spec=SceptreContext)
        self.mock_config_reader = MagicMock(spec=ConfigReader)
        self.mock_context.project_path = sentinel.project_path
        self.mock_context.command_path = sentinel.command_path
        self.mock_context.config_file = sentinel.config_file
        self.mock_context.full_config_path.return_value =\
            sentinel.full_config_path
        self.mock_context.user_variables = {}
        self.mock_context.options = {}
        self.mock_context.no_colour = True
        self.mock_config_reader.context = self.mock_context

    def test_planner_executes_without_params(self):
        plan = MagicMock(spec=SceptrePlan)
        plan.context = self.mock_context
        plan.launch.return_value = sentinel.success
        result = plan.launch()
        plan.launch.assert_called_once_with()
        assert result == sentinel.success

    def test_planner_executes_with_params(self):
        plan = MagicMock(spec=SceptrePlan)
        plan.context = self.mock_context
        plan.launch.return_value = sentinel.success
        result = plan.launch('test-attribute')
        plan.launch.assert_called_once_with('test-attribute')
        assert result == sentinel.success

    def test_command_not_found_error_raised(self):
        with pytest.raises(AttributeError):
            plan = MagicMock(spec=SceptrePlan)
            plan.context = self.mock_context
            plan.invalid_command()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os
from setuptools import setup, find_packages


def read_file(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read_file(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


install_requirements = [
    "boto3>=1.3,<2.0",
    "click>=7.0,<9.0",
    "cfn-flip>=1.2.3,<2.0",
    "deepdiff>=5.5.0,<6.0",
    "PyYaml>6.0,<7.0",
    "Jinja2>=3.0,<4",
    "jsonschema>=3.2,<3.3",
    "colorama>=0.3.9",
    "packaging>=16.8,<22.0",
    "sceptre-cmd-resolver>=1.1.3,<2",
    "sceptre-file-resolver>=1.0.4,<2",
    "six>=1.11.0,<2.0.0",
    "networkx>=2.6,<2.7",
    "urllib3<2.0"
]

extra_requirements = {
    "troposphere": ["troposphere>=4,<5"],
}


setup(
    name="sceptre",
    version=get_version("sceptre/__init__.py"),
    description="Cloud Provisioning Tool",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    author="Cloudreach",
    author_email="sceptre@cloudreach.com",
    license='Apache2',
    url="https://github.com/Sceptre/sceptre",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    package_dir={
        "sceptre": "sceptre"
    },
    py_modules=["sceptre"],
    entry_points={
        "console_scripts": [
            'sceptre = sceptre.cli:cli'
        ],
        "sceptre.hooks": [
            "asg_scheduled_actions ="
            "sceptre.hooks.asg_scaling_processes:ASGScalingProcesses",
            "cmd = sceptre.hooks.cmd:Cmd"
        ],
        "sceptre.resolvers": [
            "environment_variable ="
            "sceptre.resolvers.environment_variable:EnvironmentVariable",
            "file_contents = sceptre.resolvers.file_contents:FileContents",
            "stack_output = sceptre.resolvers.stack_output:StackOutput",
            "stack_output_external ="
            "sceptre.resolvers.stack_output:StackOutputExternal",
            "no_value = sceptre.resolvers.no_value:NoValue",
            "stack_attr = sceptre.resolvers.stack_attr:StackAttr"
        ],
        "sceptre.template_handlers": [
            "file = sceptre.template_handlers.file:File",
            "s3 = sceptre.template_handlers.s3:S3",
            "http = sceptre.template_handlers.http:Http"
        ]
    },
    data_files=[
        (os.path.join("sceptre", "stack_policies"), [
            os.path.join("sceptre", "stack_policies", "lock.json"),
            os.path.join("sceptre", "stack_policies", "unlock.json")
        ])
    ],
    include_package_data=True,
    zip_safe=False,
    keywords="sceptre",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Environment :: Console",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    test_suite="tests",
    install_requires=install_requirements,
    extras_require=extra_requirements,
)



# -*- coding: utf-8 -*-

"""
sceptre.stack

This module implements a Stack class, which stores a Stack's data.

"""

import logging
from typing import List, Any

from sceptre.connection_manager import ConnectionManager
from sceptre.exceptions import InvalidConfigFileError
from sceptre.helpers import get_external_stack_name, sceptreise_path
from sceptre.hooks import Hook, HookProperty
from sceptre.resolvers import (
    ResolvableContainerProperty,
    ResolvableValueProperty,
    RecursiveResolve,
    PlaceholderType
)
from sceptre.template import Template


class Stack(object):
    """
    Stack stores information about a particular CloudFormation Stack.

    :param name: The name of the Stack.

    :param project_code: A code which is prepended to the Stack names\
            of all Stacks built by Sceptre.

    :param template_path: The relative path to the CloudFormation, Jinja2\
            or Python template to build the Stack from. If this is filled,
            `template_handler_config` should not be filled.

    :param template_handler_config: Configuration for a Template Handler that can resolve
            its arguments to a template string. Should contain the `type` property to specify
            the type of template handler to load. Conflicts with `template_path`.

    :param region: The AWS region to build Stacks in.

    :param template_bucket_name: The name of the S3 bucket the Template is uploaded to.

    :param template_key_prefix: A prefix to the key used to store templates uploaded to S3

    :param required_version: A PEP 440 compatible version specifier. If the Sceptre version does\
            not fall within the given version requirement it will abort.

    :param parameters: The keys must match up with the name of the parameter.\
            The value must be of the type as defined in the template.

    :param sceptre_user_data: Data passed into\
            `sceptre_handler(sceptre_user_data)` function in Python templates\
            or accessible under `sceptre_user_data` variable within Jinja2\
            templates.

    :param hooks: A list of arbitrary shell or python commands or scripts to\
            run.

    :param s3_details:

    :param dependencies: The relative path to the Stack, including the file\
            extension of the Stack.

    :param role_arn: The ARN of a CloudFormation Service Role that is assumed\
            by CloudFormation to create, update or delete resources.

    :param protected: Stack protection against execution.

    :param tags: CloudFormation Tags to be applied to the Stack.

    :param external_name:

    :param notifications: SNS topic ARNs to publish Stack related events to.\
            A maximum of 5 ARNs can be specified per Stack.

    :param on_failure: This parameter describes the action taken by\
            CloudFormation when a Stack fails to create.

    :param iam_role: The ARN of a role for Sceptre to assume before interacting\
            with the environment. If not supplied, Sceptre uses the user's AWS CLI\
            credentials.

    :param profile: The name of the profile as defined in ~/.aws/config and\
            ~/.aws/credentials.

    :param stack_timeout: A timeout in minutes before considering the Stack\
            deployment as failed. After the specified timeout, the Stack will\
            be rolled back. Specifying zero, as well as omitting the field,\
            will result in no timeout. Supports only positive integer value.

    :param ignore: If True, this stack will be ignored during launches (but it can be explicitly
            deployed with create, update, and delete commands.

    :param obsolete: If True, this stack will operate the same as if ignore was set, but it will
            also be deleted if the prune command is invoked or the --prune option is used with the
            launch command.

    :param iam_role_session_duration: The session duration when Scetre assumes a role.\
           If not supplied, Sceptre uses default value (3600 seconds)

    :param stack_group_config: The StackGroup config for the Stack

    """
    parameters = ResolvableContainerProperty("parameters")
    sceptre_user_data = ResolvableContainerProperty(
        "sceptre_user_data",
        PlaceholderType.alphanum
    )
    notifications = ResolvableContainerProperty("notifications")
    tags = ResolvableContainerProperty('tags')
    # placeholder_override=None here means that if the template_bucket_name is a resolver,
    # placeholders have been enabled, and that stack hasn't been deployed yet, commands that would
    # otherwise attempt to upload the template (like validate) won't actually use the template bucket
    # and will act as if there was no template bucket set.
    s3_details = ResolvableContainerProperty(
        "s3_details",
        PlaceholderType.none
    )
    template_handler_config = ResolvableContainerProperty(
        'template_handler_config',
        PlaceholderType.alphanum
    )

    template_bucket_name = ResolvableValueProperty(
        "template_bucket_name",
        PlaceholderType.none
    )
    # Similarly, the placeholder_override=None for iam_role means that actions that would otherwise
    # use the iam_role will act as if there was no iam role when the iam_role stack has not been
    # deployed for commands that allow placeholders (like validate).
    iam_role = ResolvableValueProperty(
        'iam_role',
        PlaceholderType.none
    )
    role_arn = ResolvableValueProperty('role_arn')

    hooks = HookProperty("hooks")

    def __init__(
        self, name: str, project_code: str, region: str, template_path: str = None,
        template_handler_config: dict = None, template_bucket_name: str = None, template_key_prefix: str = None,
        required_version: str = None, parameters: dict = None, sceptre_user_data: dict = None, hooks: Hook = None,
        s3_details: dict = None, iam_role: str = None, dependencies: List["Stack"] = None, role_arn: str = None,
        protected: bool = False, tags: dict = None, external_name: str = None, notifications: List[str] = None,
        on_failure: str = None, profile: str = None, stack_timeout: int = 0, iam_role_session_duration: int = 0,
        ignore=False, obsolete=False, stack_group_config: dict = {}
    ):
        self.logger = logging.getLogger(__name__)

        if template_path and template_handler_config:
            raise InvalidConfigFileError("Both 'template_path' and 'template' are set, specify one or the other")

        if not template_path and not template_handler_config:
            raise InvalidConfigFileError("Neither 'template_path' nor 'template' is set")

        self.name = sceptreise_path(name)
        self.project_code = project_code
        self.region = region
        self.required_version = required_version
        self.external_name = external_name or get_external_stack_name(self.project_code, self.name)
        self.template_path = template_path
        self.dependencies = dependencies or []
        self.protected = protected
        self.on_failure = on_failure
        self.stack_group_config = stack_group_config or {}
        self.stack_timeout = stack_timeout
        self.profile = profile
        self.template_key_prefix = template_key_prefix
        self.iam_role_session_duration = iam_role_session_duration
        self.ignore = self._ensure_boolean("ignore", ignore)
        self.obsolete = self._ensure_boolean("obsolete", obsolete)

        self._template = None
        self._connection_manager = None

        # Resolvers and hooks need to be assigned last
        self.s3_details = s3_details
        self.iam_role = iam_role
        self.tags = tags or {}
        self.role_arn = role_arn
        self.template_bucket_name = template_bucket_name
        self.template_handler_config = template_handler_config

        self.s3_details = s3_details
        self.parameters = parameters or {}
        self.sceptre_user_data = sceptre_user_data or {}
        self.notifications = notifications or []

        self.hooks = hooks or {}

    def _ensure_boolean(self, config_name: str, value: Any) -> bool:
        if not isinstance(value, bool):
            raise InvalidConfigFileError(
                f"{self.name}: Value for {config_name} must be a boolean, not a {type(value).__name__}"
            )
        return value

    def __repr__(self):
        return (
            "sceptre.stack.Stack("
            f"name='{self.name}', "
            f"project_code={self.project_code}, "
            f"template_path={self.template_path}, "
            f"template_handler_config={self.template_handler_config}, "
            f"region={self.region}, "
            f"template_bucket_name={self.template_bucket_name}, "
            f"template_key_prefix={self.template_key_prefix}, "
            f"required_version={self.required_version}, "
            f"iam_role={self.iam_role}, "
            f"iam_role_session_duration={self.iam_role_session_duration}, "
            f"profile={self.profile}, "
            f"sceptre_user_data={self.sceptre_user_data}, "
            f"parameters={self.parameters}, "
            f"hooks={self.hooks}, "
            f"s3_details={self.s3_details}, "
            f"dependencies={self.dependencies}, "
            f"role_arn={self.role_arn}, "
            f"protected={self.protected}, "
            f"tags={self.tags}, "
            f"external_name={self.external_name}, "
            f"notifications={self.notifications}, "
            f"on_failure={self.on_failure}, "
            f"stack_timeout={self.stack_timeout}, "
            f"stack_group_config={self.stack_group_config}, "
            f"ignore={self.ignore}, "
            f"obsolete={self.obsolete}"
            ")"
        )

    def __str__(self):
        return self.name

    def __eq__(self, stack):
        # We should not use any resolvable properties in __eq__, since it is used when adding the
        # Stack to a set, which is done very early in plan resolution. Trying to reference resolvers
        # before the plan is fully resolved can potentially blow up.
        return (
            self.name == stack.name and
            self.external_name == stack.external_name and
            self.project_code == stack.project_code and
            self.template_path == stack.template_path and
            self.region == stack.region and
            self.template_key_prefix == stack.template_key_prefix and
            self.required_version == stack.required_version and
            self.iam_role_session_duration == stack.iam_role_session_duration and
            self.profile == stack.profile and
            self.dependencies == stack.dependencies and
            self.protected == stack.protected and
            self.on_failure == stack.on_failure and
            self.stack_timeout == stack.stack_timeout and
            self.ignore == stack.ignore and
            self.obsolete == stack.obsolete
        )

    def __hash__(self):
        return hash(str(self))

    @property
    def connection_manager(self) -> ConnectionManager:
        """Returns the ConnectionManager for the stack, creating it if it has not yet been created.

        :returns: ConnectionManager.
        """
        if self._connection_manager is None:
            cache_connection_manager = True
            try:
                iam_role = self.iam_role
            except RecursiveResolve:
                # This would be the case when iam_role is set with a resolver (especially stack_output)
                # that uses the stack's connection manager. This creates a temporary condition where
                # you need the iam role to get the iam role. To get around this, it will temporarily
                # use None as the iam_role but will re-attempt to resolve the value in future accesses.
                # Since the Stack Output resolver (the most likely culprit) uses the target stack's
                # iam_role rather than the current stack's one anyway, it actually doesn't matter,
                # since the stack defining that iam_role won't actually be using that iam_role.
                self.logger.debug(
                    "Resolving iam_role requires the Stack connection manager. Temporarily setting "
                    "the iam_role to None until it can be fully resolved."
                )
                iam_role = None
                cache_connection_manager = False

            connection_manager = ConnectionManager(
                self.region, self.profile, self.external_name, iam_role, self.iam_role_session_duration
            )
            if cache_connection_manager:
                self._connection_manager = connection_manager
            else:  # Return early without caching the connection manager.
                return connection_manager

        return self._connection_manager

    @property
    def template(self):
        """
        Returns the CloudFormation Template used to create the Stack.

        :returns: The Stack's template.
        :rtype: Template
        """
        if self._template is None:
            if self.template_path:
                handler_config = {
                    "type": "file",
                    "path": self.template_path
                }
            else:
                handler_config = self.template_handler_config

            self._template = Template(
                name=self.name,
                handler_config=handler_config,
                sceptre_user_data=self.sceptre_user_data,
                stack_group_config=self.stack_group_config,
                s3_details=self.s3_details,
                connection_manager=self.connection_manager
            )
        return self._template



# -*- coding: utf-8 -*-

"""
sceptre.plan.actions

This module implements the StackActions class which provides the functionality
available to a Stack.
"""

import json
import logging
import time
import typing
import urllib
from datetime import datetime, timedelta
from os import path
from typing import Union, Optional, Tuple, Dict

import botocore
from dateutil.tz import tzutc

from sceptre.config.reader import ConfigReader
from sceptre.connection_manager import ConnectionManager
from sceptre.exceptions import (
    CannotUpdateFailedStackError,
    ProtectedStackError,
    StackDoesNotExistError,
    UnknownStackChangeSetStatusError,
    UnknownStackStatusError
)
from sceptre.helpers import normalise_path
from sceptre.hooks import add_stack_hooks
from sceptre.stack import Stack
from sceptre.stack_status import StackChangeSetStatus, StackStatus

if typing.TYPE_CHECKING:
    from sceptre.diffing.stack_differ import StackDiff, StackDiffer


class StackActions(object):
    """
    StackActions stores the operations a Stack can take, such as creating or
    deleting the Stack.

    :param stack: A Stack object
    :type stack: sceptre.stack.Stack
    """

    def __init__(self, stack: Stack):
        self.stack = stack
        self.name = self.stack.name
        self.logger = logging.getLogger(__name__)
        self.connection_manager = ConnectionManager(
            self.stack.region, self.stack.profile,
            self.stack.external_name, self.stack.iam_role,
            self.stack.iam_role_session_duration
        )

    @add_stack_hooks
    def create(self):
        """
        Creates a Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()
        self.logger.info("%s - Creating Stack", self.stack.name)
        create_stack_kwargs = {
            "StackName": self.stack.external_name,
            "Parameters": self._format_parameters(self.stack.parameters),
            "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM', 'CAPABILITY_AUTO_EXPAND'],
            "NotificationARNs": self.stack.notifications,
            "Tags": [
                {"Key": str(k), "Value": str(v)}
                for k, v in self.stack.tags.items()
            ]
        }

        if self.stack.on_failure:
            create_stack_kwargs.update({"OnFailure": self.stack.on_failure})
        create_stack_kwargs.update(
            self.stack.template.get_boto_call_parameter())
        create_stack_kwargs.update(self._get_role_arn())
        create_stack_kwargs.update(self._get_stack_timeout())

        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="create_stack",
                kwargs=create_stack_kwargs
            )

            self.logger.debug(
                "%s - Create stack response: %s", self.stack.name, response
            )

            status = self._wait_for_completion()
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Code"] == "AlreadyExistsException":
                self.logger.info(
                    "%s - Stack already exists", self.stack.name
                )

                status = StackStatus.COMPLETE
            else:
                raise

        return status

    @add_stack_hooks
    def update(self):
        """
        Updates the Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()
        self.logger.info("%s - Updating Stack", self.stack.name)
        try:
            update_stack_kwargs = {
                "StackName": self.stack.external_name,
                "Parameters": self._format_parameters(self.stack.parameters),
                "Capabilities": [
                    'CAPABILITY_IAM',
                    'CAPABILITY_NAMED_IAM',
                    'CAPABILITY_AUTO_EXPAND'
                ],
                "NotificationARNs": self.stack.notifications,
                "Tags": [
                    {"Key": str(k), "Value": str(v)}
                    for k, v in self.stack.tags.items()
                ]
            }
            update_stack_kwargs.update(
                self.stack.template.get_boto_call_parameter())
            update_stack_kwargs.update(self._get_role_arn())
            response = self.connection_manager.call(
                service="cloudformation",
                command="update_stack",
                kwargs=update_stack_kwargs
            )
            status = self._wait_for_completion(self.stack.stack_timeout)
            self.logger.debug(
                "%s - Update Stack response: %s", self.stack.name, response
            )

            # Cancel update after timeout
            if status == StackStatus.IN_PROGRESS:
                status = self.cancel_stack_update()

            return status
        except botocore.exceptions.ClientError as exp:
            error_message = exp.response["Error"]["Message"]
            if error_message == "No updates are to be performed.":
                self.logger.info(
                    "%s - No updates to perform.", self.stack.name
                )
                return StackStatus.COMPLETE
            else:
                raise

    def cancel_stack_update(self):
        """
        Cancels a Stack update.

        :returns: The cancelled Stack status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self.logger.warning(
            "%s - Update Stack time exceeded the specified timeout",
            self.stack.name
        )
        response = self.connection_manager.call(
            service="cloudformation",
            command="cancel_update_stack",
            kwargs={"StackName": self.stack.external_name}
        )
        self.logger.debug(
            "%s - Cancel update Stack response: %s", self.stack.name, response
        )
        return self._wait_for_completion()

    @add_stack_hooks
    def launch(self) -> StackStatus:
        """
        Launches the Stack.

        If the Stack status is create_failed or rollback_complete, the
        Stack is deleted. Launch then tries to create or update the Stack,
        depending if it already exists. If there are no updates to be
        performed, launch exits gracefully.

        :returns: The Stack's status.
        """
        self._protect_execution()
        self.logger.info(f"{self.stack.name} - Launching Stack")

        try:
            existing_status = self._get_status()
        except StackDoesNotExistError:
            existing_status = "PENDING"

        self.logger.info(
            "%s - Stack is in the %s state", self.stack.name, existing_status
        )

        if existing_status == "PENDING":
            status = self.create()
        elif existing_status in ["CREATE_FAILED", "ROLLBACK_COMPLETE"]:
            self.delete()
            status = self.create()
        elif existing_status.endswith("COMPLETE"):
            status = self.update()
        elif existing_status.endswith("IN_PROGRESS"):
            self.logger.info(
                "%s - Stack action is already in progress state and cannot "
                "be updated", self.stack.name
            )
            status = StackStatus.IN_PROGRESS
        elif existing_status.endswith("FAILED"):
            raise CannotUpdateFailedStackError(
                "'{0}' is in a the state '{1}' and cannot be updated".format(
                    self.stack.name, existing_status
                )
            )
        else:
            raise UnknownStackStatusError(
                "{0} is unknown".format(existing_status)
            )
        return status

    @add_stack_hooks
    def delete(self):
        """
        Deletes the Stack.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        self._protect_execution()

        self.logger.info("%s - Deleting stack", self.stack.name)
        try:
            status = self._get_status()
        except StackDoesNotExistError:
            self.logger.info("%s - Does not exist.", self.stack.name)
            status = StackStatus.COMPLETE
            return status

        delete_stack_kwargs = {"StackName": self.stack.external_name}
        delete_stack_kwargs.update(self._get_role_arn())
        self.connection_manager.call(
            service="cloudformation",
            command="delete_stack",
            kwargs=delete_stack_kwargs
        )

        try:
            status = self._wait_for_completion()
        except StackDoesNotExistError:
            status = StackStatus.COMPLETE
        except botocore.exceptions.ClientError as error:
            if error.response["Error"]["Message"].endswith("does not exist"):
                status = StackStatus.COMPLETE
            else:
                raise
        self.logger.info("%s - delete %s", self.stack.name, status)
        return status

    def lock(self):
        """
        Locks the Stack by applying a deny-all updates Stack Policy.
        """
        policy_path = path.join(
            # need to get to the base install path. __file__ will take us into
            # sceptre/actions so need to walk up the path.
            path.abspath(path.join(__file__, "..", "..")),
            "stack_policies/lock.json"
        )
        self.set_policy(policy_path)
        self.logger.info("%s - Successfully locked Stack", self.stack.name)

    def unlock(self):
        """
        Unlocks the Stack by applying an allow-all updates Stack Policy.
        """
        policy_path = path.join(
            # need to get to the base install path. __file__ will take us into
            # sceptre/actions so need to walk up the path.
            path.abspath(path.join(__file__, "..", "..")),
            "stack_policies/unlock.json"
        )
        self.set_policy(policy_path)
        self.logger.info("%s - Successfully unlocked Stack", self.stack.name)

    def describe(self):
        """
        Returns the a description of the Stack.

        :returns: A Stack description.
        :rtype: dict
        """
        try:
            return self.connection_manager.call(
                service="cloudformation",
                command="describe_stacks",
                kwargs={"StackName": self.stack.external_name}
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Message"].endswith("does not exist"):
                return
            raise

    def describe_events(self):
        """
        Returns the CloudFormation events for a Stack.

        :returns: CloudFormation events for a Stack.
        :rtype: dict
        """
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_events",
            kwargs={"StackName": self.stack.external_name}
        )

    def describe_resources(self):
        """
        Returns the logical and physical resource IDs of the Stack's resources.

        :returns: Information about the Stack's resources.
        :rtype: dict
        """
        self.logger.debug("%s - Describing stack resources", self.stack.name)
        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="describe_stack_resources",
                kwargs={"StackName": self.stack.external_name}
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Message"].endswith("does not exist"):
                return {self.stack.name: []}
            raise

        self.logger.debug(
            "%s - Describe Stack resource response: %s",
            self.stack.name,
            response
        )

        desired_properties = ["LogicalResourceId", "PhysicalResourceId"]

        formatted_response = {self.stack.name: [
            {k: v for k, v in item.items() if k in desired_properties}
            for item in response["StackResources"]
        ]}
        return formatted_response

    def describe_outputs(self):
        """
        Returns the Stack's outputs.

        :returns: The Stack's outputs.
        :rtype: list
        """
        self.logger.debug("%s - Describing stack outputs", self.stack.name)

        try:
            response = self._describe()
        except botocore.exceptions.ClientError:
            return []

        return {self.stack.name: response["Stacks"][0].get("Outputs", [])}

    def continue_update_rollback(self):
        """
        Rolls back a Stack in the UPDATE_ROLLBACK_FAILED state to
        UPDATE_ROLLBACK_COMPLETE.
        """
        self.logger.debug("%s - Continuing update rollback", self.stack.name)
        continue_update_rollback_kwargs = {
            "StackName": self.stack.external_name
        }
        continue_update_rollback_kwargs.update(self._get_role_arn())
        self.connection_manager.call(
            service="cloudformation",
            command="continue_update_rollback",
            kwargs=continue_update_rollback_kwargs
        )
        self.logger.info(
            "%s - Successfully initiated continuation of update rollback",
            self.stack.name
        )

    def set_policy(self, policy_path):
        """
        Applies a Stack Policy.

        :param policy_path: The relative path of JSON file containing\
                the AWS Policy to apply.
        :type policy_path: str
        """
        with open(policy_path) as f:
            policy = f.read()

        self.logger.debug(
            "%s - Setting Stack policy: \n%s",
            self.stack.name,
            policy
        )

        self.connection_manager.call(
            service="cloudformation",
            command="set_stack_policy",
            kwargs={
                "StackName": self.stack.external_name,
                "StackPolicyBody": policy
            }
        )
        self.logger.info("%s - Successfully set Stack Policy", self.stack.name)

    def get_policy(self):
        """
        Returns a Stack's Policy.

        :returns: The Stack's Stack Policy.
        :rtype: str
        """
        self.logger.debug("%s - Getting Stack Policy", self.stack.name)
        response = self.connection_manager.call(
            service="cloudformation",
            command="get_stack_policy",
            kwargs={
                "StackName": self.stack.external_name
            }
        )
        json_formatting = json.loads(response.get(
            "StackPolicyBody", json.dumps("No Policy Information")))
        return {self.stack.name: json_formatting}

    @add_stack_hooks
    def create_change_set(self, change_set_name):
        """
        Creates a Change Set with the name ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        """
        create_change_set_kwargs = {
            "StackName": self.stack.external_name,
            "Parameters": self._format_parameters(self.stack.parameters),
            "Capabilities": ['CAPABILITY_IAM', 'CAPABILITY_NAMED_IAM', 'CAPABILITY_AUTO_EXPAND'],
            "ChangeSetName": change_set_name,
            "NotificationARNs": self.stack.notifications,
            "Tags": [
                {"Key": str(k), "Value": str(v)}
                for k, v in self.stack.tags.items()
            ]
        }
        create_change_set_kwargs.update(
            self.stack.template.get_boto_call_parameter()
        )
        create_change_set_kwargs.update(self._get_role_arn())
        self.logger.debug(
            "%s - Creating Change Set '%s'", self.stack.name, change_set_name
        )
        self.connection_manager.call(
            service="cloudformation",
            command="create_change_set",
            kwargs=create_change_set_kwargs
        )
        # After the call successfully completes, AWS CloudFormation
        # starts creating the Change Set.
        self.logger.info(
            "%s - Successfully initiated creation of Change Set '%s'",
            self.stack.name, change_set_name
        )

    def delete_change_set(self, change_set_name):
        """
        Deletes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        """
        self.logger.debug(
            "%s - Deleting Change Set '%s'", self.stack.name, change_set_name
        )
        self.connection_manager.call(
            service="cloudformation",
            command="delete_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name
            }
        )
        # If the call successfully completes, AWS CloudFormation
        # successfully deleted the Change Set.
        self.logger.info(
            "%s - Successfully deleted Change Set '%s'",
            self.stack.name, change_set_name
        )

    def describe_change_set(self, change_set_name):
        """
        Describes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The description of the Change Set.
        :rtype: dict
        """
        self.logger.debug(
            "%s - Describing Change Set '%s'", self.stack.name, change_set_name
        )
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name
            }
        )

    def execute_change_set(self, change_set_name):
        """
        Executes the Change Set ``change_set_name``.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Stack status
        :rtype: str
        """
        self._protect_execution()
        change_set = self.describe_change_set(change_set_name)
        status = change_set.get("Status")
        reason = change_set.get("StatusReason")
        if status == "FAILED" and self.change_set_creation_failed_due_to_no_changes(reason):
            self.logger.info(
                    "Skipping ChangeSet on Stack: {} - there are no changes".format(
                        change_set.get("StackName")
                        )
            )
            return 0

        self.logger.debug(
            "%s - Executing Change Set '%s'", self.stack.name, change_set_name
        )
        self.connection_manager.call(
            service="cloudformation",
            command="execute_change_set",
            kwargs={
                "ChangeSetName": change_set_name,
                "StackName": self.stack.external_name
            }
        )

        status = self._wait_for_completion()
        return status

    def change_set_creation_failed_due_to_no_changes(self, reason: str) -> bool:
        """Indicates the change set failed when it was created because there were actually
        no changes introduced from the change set.

        :param reason: The reason reported by CloudFormation for the Change Set failure
        """
        reason = reason.lower()
        no_change_substrings = (
            "submitted information didn't contain changes",
            "no updates are to be performed"  # The reason returned for SAM templates
        )

        for substring in no_change_substrings:
            if substring in reason:
                return True
        return False

    def list_change_sets(self, url=False):
        """
        Lists the Stack's Change Sets.

        :param url: Write out a console URL instead.
        :type url: bool

        :returns: The Stack's Change Sets.
        :rtype: dict or list
        """
        response = self._list_change_sets()
        summaries = response.get("Summaries", [])

        if url:
            summaries = self._convert_to_url(summaries)

        return {self.stack.name: summaries}

    def _list_change_sets(self):
        self.logger.debug("%s - Listing change sets", self.stack.name)
        try:
            return self.connection_manager.call(
                service="cloudformation",
                command="list_change_sets",
                kwargs={
                    "StackName": self.stack.external_name
                }
            )
        except botocore.exceptions.ClientError:
            return []

    def _convert_to_url(self, summaries):
        """
        Convert the list_change_sets response from
        CloudFormation to a URL in the AWS Console.
        """
        new_summaries = []

        for summary in summaries:
            stack_id = summary["StackId"]
            change_set_id = summary["ChangeSetId"]

            region = self.stack.region
            encoded = urllib.parse.urlencode({
                "stackId": stack_id,
                "changeSetId": change_set_id
            })

            new_summaries.append(
                f"https://{region}.console.aws.amazon.com/cloudformation/home?"
                f"region={region}#/stacks/changesets/changes?{encoded}"
            )

        return new_summaries

    @add_stack_hooks
    def generate(self):
        """
        Returns the Template for the Stack
        """
        return self.stack.template.body

    @add_stack_hooks
    def validate(self):
        """
        Validates the Stack's CloudFormation Template.

        Raises an error if the Template is invalid.

        :returns: Validation information about the Template.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.logger.debug("%s - Validating Template", self.stack.name)
        response = self.connection_manager.call(
            service="cloudformation",
            command="validate_template",
            kwargs=self.stack.template.get_boto_call_parameter()
        )
        self.logger.debug(
            "%s - Validate Template response: %s", self.stack.name, response
        )
        return response

    def estimate_cost(self):
        """
        Estimates a Stack's cost.

        :returns: An estimate of the Stack's cost.
        :rtype: dict
        :raises: botocore.exceptions.ClientError
        """
        self.logger.debug("%s - Estimating template cost", self.stack.name)

        parameters = [
            {'ParameterKey': key, 'ParameterValue': value}
            for key, value in self.stack.parameters.items()
        ]

        kwargs = self.stack.template.get_boto_call_parameter()
        kwargs.update({'Parameters': parameters})

        response = self.connection_manager.call(
            service="cloudformation",
            command="estimate_template_cost",
            kwargs=kwargs
        )
        self.logger.debug(
            "%s - Estimate Stack cost response: %s", self.stack.name, response
        )
        return response

    def get_status(self):
        """
        Returns the Stack's status.

        :returns: The Stack's status.
        :rtype: sceptre.stack_status.StackStatus
        """
        try:
            return self._get_status()
        except StackDoesNotExistError:
            return "PENDING"

    def _format_parameters(self, parameters):
        """
        Converts CloudFormation parameters to the format used by Boto3.

        :param parameters: A dictionary of parameters.
        :type parameters: dict
        :returns: A list of the formatted parameters.
        :rtype: list
        """
        formatted_parameters = []
        for name, value in parameters.items():
            if value is None:
                continue
            if isinstance(value, list):
                value = ",".join(value)
            formatted_parameters.append({
                "ParameterKey": name,
                "ParameterValue": value
            })

        return formatted_parameters

    def _get_role_arn(self):
        """
        Returns the Role ARN assumed by CloudFormation when building a Stack.

        Returns an empty dict if no Role is to be assumed.

        :returns: The a Role ARN
        :rtype: dict
        """
        if self.stack.role_arn:
            return {
                "RoleARN": self.stack.role_arn
            }
        else:
            return {}

    def _get_stack_timeout(self):
        """
        Return the timeout before considering the Stack to be failing.

        Returns an empty dict if no timeout is set.
        :returns: the creation/update timeout
        :rtype: dict
        """
        if self.stack.stack_timeout:
            return {
                "TimeoutInMinutes": self.stack.stack_timeout
            }
        else:
            return {}

    def _protect_execution(self):
        """
        Raises a ProtectedStackError if protect == True.

        :raises: sceptre.exceptions.ProtectedStackError
        """
        if self.stack.protected:
            raise ProtectedStackError(
                "Cannot perform action on '{0}': Stack protection is "
                "currently enabled".format(self.stack.name)
            )

    def _wait_for_completion(self, timeout=0):
        """
        Waits for a Stack operation to finish. Prints CloudFormation events
        while it waits.

        :param timeout: Timeout before returning, in minutes.

        :returns: The final Stack status.
        :rtype: sceptre.stack_status.StackStatus
        """
        timeout = 60 * timeout

        def timed_out(elapsed):
            return elapsed >= timeout if timeout else False

        status = StackStatus.IN_PROGRESS

        self.most_recent_event_datetime = (
            datetime.now(tzutc()) - timedelta(seconds=3)
        )
        elapsed = 0
        while status == StackStatus.IN_PROGRESS and not timed_out(elapsed):
            status = self._get_simplified_status(self._get_status())
            self._log_new_events()
            time.sleep(4)
            elapsed += 4

        return status

    def _describe(self):
        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stacks",
            kwargs={"StackName": self.stack.external_name}
        )

    def _get_status(self):
        try:
            status = self._describe()["Stacks"][0]["StackStatus"]
        except botocore.exceptions.ClientError as exp:
            if exp.response["Error"]["Message"].endswith("does not exist"):
                raise StackDoesNotExistError(exp.response["Error"]["Message"])
            else:
                raise exp
        return status

    @staticmethod
    def _get_simplified_status(status):
        """
        Returns the simplified Stack Status.

        The simplified Stack status is represented by the struct
        ``sceptre.StackStatus()`` and can take one of the following options:

        * complete
        * in_progress
        * failed

        :param status: The CloudFormation Stack status to simplify.
        :type status: str
        :returns: The Stack's simplified status
        :rtype: sceptre.stack_status.StackStatus
        """
        if status.endswith("ROLLBACK_COMPLETE"):
            return StackStatus.FAILED
        elif status.endswith("_COMPLETE"):
            return StackStatus.COMPLETE
        elif status.endswith("_IN_PROGRESS"):
            return StackStatus.IN_PROGRESS
        elif status.endswith("_FAILED"):
            return StackStatus.FAILED
        else:
            raise UnknownStackStatusError(
                "{0} is unknown".format(status)
            )

    def _log_new_events(self):
        """
        Log the latest Stack events while the Stack is being built.
        """
        events = self.describe_events()["StackEvents"]
        events.reverse()
        new_events = [
            event for event in events
            if event["Timestamp"] > self.most_recent_event_datetime
        ]
        for event in new_events:
            self.logger.info(" ".join([
                self.stack.name,
                event["LogicalResourceId"],
                event["ResourceType"],
                event["ResourceStatus"],
                event.get("ResourceStatusReason", "")
            ]))
            self.most_recent_event_datetime = event["Timestamp"]

    def wait_for_cs_completion(self, change_set_name):
        """
        Waits while the Stack Change Set status is "pending".

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Change Set's status.
        :rtype: sceptre.stack_status.StackChangeSetStatus
        """
        while True:
            status = self._get_cs_status(change_set_name)
            if status != StackChangeSetStatus.PENDING:
                break
            time.sleep(2)

        return status

    def _get_cs_status(self, change_set_name):
        """
        Returns the status of a Change Set.

        :param change_set_name: The name of the Change Set.
        :type change_set_name: str
        :returns: The Change Set's status.
        :rtype: sceptre.stack_status.StackChangeSetStatus
        """
        cs_description = self.describe_change_set(change_set_name)

        cs_status = cs_description["Status"]
        cs_exec_status = cs_description["ExecutionStatus"]
        possible_statuses = [
            "CREATE_PENDING", "CREATE_IN_PROGRESS",
            "CREATE_COMPLETE", "DELETE_COMPLETE", "FAILED"
        ]
        possible_execution_statuses = [
            "UNAVAILABLE", "AVAILABLE", "EXECUTE_IN_PROGRESS",
            "EXECUTE_COMPLETE", "EXECUTE_FAILED", "OBSOLETE"
        ]

        if cs_status not in possible_statuses:
            raise UnknownStackChangeSetStatusError(
                "Status {0} is unknown".format(cs_status)
            )
        if cs_exec_status not in possible_execution_statuses:
            raise UnknownStackChangeSetStatusError(
                "ExecutionStatus {0} is unknown".format(cs_status)
            )

        if (
                cs_status == "CREATE_COMPLETE" and
                cs_exec_status == "AVAILABLE"
        ):
            return StackChangeSetStatus.READY
        elif (
                cs_status in [
                    "CREATE_PENDING", "CREATE_IN_PROGRESS", "CREATE_COMPLETE"
                ] and
                cs_exec_status in ["UNAVAILABLE", "AVAILABLE"]
        ):
            return StackChangeSetStatus.PENDING
        elif (
                cs_status in ["DELETE_COMPLETE", "FAILED"] or
                cs_exec_status in [
                    "EXECUTE_IN_PROGRESS", "EXECUTE_COMPLETE",
                    "EXECUTE_FAILED", "OBSOLETE"
                ]
        ):
            return StackChangeSetStatus.DEFUNCT
        else:  # pragma: no cover
            raise Exception("This else should not be reachable.")

    def fetch_remote_template(self) -> Optional[str]:
        """
        Returns the Template for the remote Stack

        :returns: the template body.
        """
        self.logger.debug(f"{self.stack.name} - Fetching remote template")

        original_template = self._fetch_original_template_stage()

        if isinstance(original_template, dict):
            # While not documented behavior, boto3 will attempt to deserialize the TemplateBody
            # with json.loads and return the template as a dict if it is successful; otherwise (such
            # as in when the template is in yaml, it will return the string. Therefore, we need to
            # dump the template to json if we get a dict.
            original_template = json.dumps(original_template, indent=4)

        return original_template

    def _fetch_original_template_stage(self) -> Optional[Union[str, dict]]:
        try:
            response = self.connection_manager.call(
                service="cloudformation",
                command="get_template",
                kwargs={
                    "StackName": self.stack.external_name,
                    "TemplateStage": 'Original'
                }
            )
            return response['TemplateBody']
            # Sometimes boto returns a string, sometimes a dictionary
        except botocore.exceptions.ClientError as e:
            # AWS returns a ValidationError if the stack doesn't exist
            if e.response['Error']['Code'] == 'ValidationError':
                return None
            raise

    def fetch_remote_template_summary(self):
        return self._get_template_summary(StackName=self.stack.external_name)

    def fetch_local_template_summary(self):
        boto_call_parameter = self.stack.template.get_boto_call_parameter()
        return self._get_template_summary(**boto_call_parameter)

    def _get_template_summary(self, **kwargs) -> Optional[dict]:
        try:
            template_summary = self.connection_manager.call(
                service='cloudformation',
                command='get_template_summary',
                kwargs=kwargs
            )
            return template_summary
        except botocore.exceptions.ClientError as e:
            error_response = e.response['Error']
            if (
                error_response['Code'] == 'ValidationError'
                and 'does not exist' in error_response['Message']
            ):
                return None
            raise

    @add_stack_hooks
    def diff(self, stack_differ: "StackDiffer") -> "StackDiff":
        """
        Returns a diff of local and deployed template and stack configuration using a specific diff
        library.

        :param stack_differ: The differ to use
        :returns: A StackDiff object with the full, computed diff
        """
        return stack_differ.diff(self)

    @add_stack_hooks
    def drift_detect(self) -> Dict[str, str]:
        """
        Show stack drift for a running stack.

        :returns: The stack drift detection status.
            If the stack does not exist, we return a detection and
            stack drift status of STACK_DOES_NOT_EXIST.
            If drift detection times out after 5 minutes, we return
            TIMED_OUT.
        """
        try:
            self._get_status()
        except StackDoesNotExistError:
            self.logger.info(f"{self.stack.name} - Does not exist.")
            return {
                "DetectionStatus": "STACK_DOES_NOT_EXIST",
                "StackDriftStatus": "STACK_DOES_NOT_EXIST"
            }

        response = self._detect_stack_drift()
        detection_id = response["StackDriftDetectionId"]

        try:
            response = self._wait_for_drift_status(detection_id)
        except TimeoutError as exc:
            self.logger.info(f"{self.stack.name} - {exc}")
            response = {
                "DetectionStatus": "TIMED_OUT",
                "StackDriftStatus": "TIMED_OUT"
            }

        return response

    @add_stack_hooks
    def drift_show(self, drifted: bool = False) -> Tuple[str, dict]:
        """
        Detect drift status on stacks.

        :param drifted: Filter out IN_SYNC resources.
        :returns: The detection status and resource drifts.
        """
        response = self.drift_detect()
        detection_status = response["DetectionStatus"]

        if detection_status in ["DETECTION_COMPLETE", "DETECTION_FAILED"]:
            response = self._describe_stack_resource_drifts()
        elif detection_status in ["TIMED_OUT", "STACK_DOES_NOT_EXIST"]:
            response = {"StackResourceDriftStatus": detection_status}
        else:
            raise Exception("Not expected to be reachable")

        response = self._filter_drifts(response, drifted)
        return (detection_status, response)

    def _wait_for_drift_status(self, detection_id: str) -> dict:
        """
        Waits for drift detection to complete.

        :param detection_id: The drift detection ID.
        :returns: The response from describe_stack_drift_detection_status.
        """
        timeout = 300
        sleep_interval = 10
        elapsed = 0

        while True:
            if elapsed >= timeout:
                raise TimeoutError(f"Timed out after {elapsed} seconds")

            self.logger.info(f"{self.stack.name} - Waiting for drift detection")
            response = self._describe_stack_drift_detection_status(detection_id)
            detection_status = response["DetectionStatus"]

            self._log_drift_status(response)

            if detection_status == "DETECTION_IN_PROGRESS":
                time.sleep(sleep_interval)
                elapsed += sleep_interval
            else:
                return response

    def _log_drift_status(self, response: dict) -> None:
        """
        Log the drift status while waiting for
        drift detection to complete.
        """
        keys = [
            "StackDriftDetectionId",
            "DetectionStatus",
            "DetectionStatusReason",
            "StackDriftStatus"
        ]

        for key in keys:
            if key in response:
                self.logger.debug(
                    f"{self.stack.name} - {key} - {response[key]}"
                )

    def _detect_stack_drift(self) -> dict:
        """
        Run detect_stack_drift.
        """
        self.logger.info(f"{self.stack.name} - Detecting Stack Drift")

        return self.connection_manager.call(
            service="cloudformation",
            command="detect_stack_drift",
            kwargs={
                "StackName": self.stack.external_name
            }
        )

    def _describe_stack_drift_detection_status(self, detection_id: str) -> dict:
        """
        Run describe_stack_drift_detection_status.
        """
        self.logger.info(f"{self.stack.name} - Describing Stack Drift Detection Status")

        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_drift_detection_status",
            kwargs={
                "StackDriftDetectionId": detection_id
            }
        )

    def _describe_stack_resource_drifts(self) -> dict:
        """
        Detects stack resource_drifts for a running stack.
        """
        self.logger.info(f"{self.stack.name} - Describing Stack Resource Drifts")

        return self.connection_manager.call(
            service="cloudformation",
            command="describe_stack_resource_drifts",
            kwargs={
                "StackName": self.stack.external_name
            }
        )

    def _filter_drifts(self, response: dict, drifted: bool) -> dict:
        """
        The filtered response after filtering out StackResourceDriftStatus.
        :param drifted: Filter out IN_SYNC resources from CLI --drifted.
        """
        if "StackResourceDrifts" not in response:
            return response

        result = {"StackResourceDrifts": []}
        include_all_drift_statuses = not drifted

        for drift in response["StackResourceDrifts"]:
            is_drifted = drift["StackResourceDriftStatus"] != "IN_SYNC"
            if include_all_drift_statuses or is_drifted:
                result["StackResourceDrifts"].append(drift)

        return result

    @add_stack_hooks
    def dump_config(self, config_reader: ConfigReader):
        """
        Dump the config for a stack.
        """
        stack_path = normalise_path(self.stack.name + ".yaml")
        return config_reader.read(stack_path)



import pytest

from sceptre.exceptions import TemplateHandlerArgumentsInvalidError
from sceptre.template_handlers import TemplateHandler


class MockTemplateHandler(TemplateHandler):
    def __init__(self, *args, **kwargs):
        super(MockTemplateHandler, self).__init__(*args, **kwargs)

    def schema(self):
        return {
            "type": "object",
            "properties": {
                "argument": {"type": "string"}
            },
            "required": ["argument"]
        }

    def handle(self):
        return "TestTemplateHandler"


class TestTemplateHandlers(object):
    def test_template_handler_validates_schema(self):
        handler = MockTemplateHandler(name="mock", arguments={"argument": "test"})
        handler.validate()

    def test_template_handler_errors_when_arguments_invalid(self):
        with pytest.raises(TemplateHandlerArgumentsInvalidError):
            handler = MockTemplateHandler(name="mock", arguments={"non-existent": "test"})
            handler.validate()



# -*- coding: utf-8 -*-

"""
sceptre.plan.executor

This module implements a SceptrePlanExecutor, which is responsible for
executing the command specified in a SceptrePlan.
"""
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import List, Set

from sceptre.plan.actions import StackActions
from sceptre.stack import Stack


class SceptrePlanExecutor(object):

    def __init__(self, command: str, launch_order: List[Set[Stack]]):
        """
        Initialises a SceptrePlanExecutor, generates the launch order, threads
        and intial Stack Statuses.

        :param command: The command to execute on the Stack.

        :param launch_order: A list containing sets of Stacks that can be executed concurrently.
        """

        self.logger = logging.getLogger(__name__)
        self.command = command
        self.launch_order = launch_order
        # Select the number of threads based upon the max batch size,
        # or use 1 if all batches are empty
        self.num_threads = len(max(launch_order, key=len)) or 1

    def execute(self, *args):
        """
        Execute is responsible executing the sets of Stacks in launch_order
        concurrently, in the correct order.

        :param args: Any arguments that should be passed through to the
                StackAction being called.
        """
        responses = {}

        with ThreadPoolExecutor(max_workers=self.num_threads) as executor:
            for batch in self.launch_order:
                futures = [executor.submit(self._execute, stack, *args)
                           for stack in batch]

                for future in as_completed(futures):
                    stack, status = future.result()
                    responses[stack] = status

        return responses

    def _execute(self, stack, *args):
        actions = StackActions(stack)
        result = getattr(actions, self.command)(*args)
        return stack, result
